import pathlib
from opengate.managers import Simulation
from opengate.geometry.volumes import (
    RepeatParametrisedVolume,
    HexagonVolume,
    unite_volumes,
)
from opengate.geometry.utility import (
    get_transform_orbiting,
)
from opengate.actors.digitizers import *
from opengate.contrib.spect.spect_helpers import (
    get_volume_position_in_head,
    get_default_energy_windows,
)
from scipy.spatial.transform import Rotation
from box import Box
from opengate.contrib.spect.spect_helpers import (
    get_mu_from_xraylib,
    calculate_acceptance_angle,
)
import json


def get_geometrical_parameters_filename():
    filename = (
        pathlib.Path(__file__).parent / "ge_discovery_nm670_geometrical_parameters.json"
    )
    return pathlib.Path(filename)


geometrical_parameters = None


def get_geometrical_parameters():
    global geometrical_parameters
    if geometrical_parameters is None:
        filename = get_geometrical_parameters_filename()
        if not filename.exists():
            print(f'update geometrical parameters to "{filename}"')
            update_geometrical_parameters(store_to_file=True)
        # print(f'Loading geometrical parameters from "{filename}"')
        with open(filename) as json_file:
            geometrical_parameters = json.load(json_file)
        geometrical_parameters = Box(geometrical_parameters)
    return geometrical_parameters


def update_geometrical_parameters(store_to_file=False):
    p = Box()
    p.collimators = ["lehr", "megp", "hegp"]
    # for all colli
    nm = g4_units.nm
    for c in p.collimators:
        s = Simulation()
        spect, colli, crystal = add_spect_head(s, "spect", c, debug=True)
        pos = get_volume_position_in_head(s, "spect", f"collimator_trd", "max", axis=2)
        y = get_volume_position_in_head(s, "spect", "crystal", "center", axis=2)
        psd = get_volume_position_in_head(s, "spect", "collimator_psd", "min", axis=2)
        p[c] = Box()
        # distance from box boundary to collimator
        p[c].collimator_position = pos
        # distance to the center of the box head
        p[c].half_box_size = spect.size[2] / 2.0 + 1 * nm
        # distance from box boundary to crystal center (for arf)
        p[c].crystal_distance = p[c].half_box_size - y
        # distance from box boundary to the collimator_psd
        p[c].psd = p[c].half_box_size - psd

        # collimator holes
        hole = s.volume_manager.get_volume(f"spect_collimator_hole")
        p[c].hole_diameter = hole.radius * 2
        p[c].collimator_length = hole.height
        holep = s.volume_manager.get_volume(f"spect_collimator_hole_param")
        tr = holep.translation
        p[c].septa_thickness = tr[1] - p[c].hole_diameter

    if store_to_file:
        filename = get_geometrical_parameters_filename()
        with open(filename, "w") as json_file:
            json.dump(p, json_file, indent=4)

    return p


def get_collimator(rad):
    radionuclides = ["tc99m", "lu177", "in111", "i131"]
    ref_collimators = ["lehr", "megp", "megp", "hegp"]
    rad = rad.lower()
    if rad not in radionuclides:
        fatal(f'The radionuclide "{rad}" is unknown. Known are: {radionuclides}')
    return ref_collimators[radionuclides.index(rad)]


def add_fake_spect_head(sim, name="spect"):
    white = [1, 1, 1, 1]
    cm = g4_units.cm
    spect_length = 19 * cm
    head = sim.add_volume("Box", name)
    head.material = "G4_AIR"
    head.size = [57.6 * cm, 44.6 * cm, spect_length]
    head.color = white
    return head


def get_orientation_for_ct(colli_type, table_shift, radius):
    nm = g4_units.nm
    pos, crystal_distance, psdd = get_plane_position_and_distance_to_crystal_OLD(
        colli_type
    )
    pos += 1 * nm
    p = [0, table_shift, -(radius + psdd)]
    return get_transform_orbiting(p, "x", 90)


def add_materials(sim):
    f = pathlib.Path(__file__).parent.resolve()
    fdb = f"{f}/spect_ge_nm670_materials.db"
    if fdb not in sim.volume_manager.material_database.filenames:
        sim.volume_manager.add_material_database(fdb)


def add_spect_head(
    sim,
    name="spect",
    collimator_type="lehr",
    rotation_deg=0,
    debug=False,
    crystal_size="3/8",
):
    """
    Collimators:
    - False : no collimator
    - lehr : holes length 35 mm, diam 1.5 mm, septal thickness : 0.2 mm
    - megp : holes length 58 mm, diam 3 mm,   septal thickness : 1.05 mm
    - hegp : holes length 66 mm, diam 4 mm,   septal thickness : 1.8 mm

    Collimator LEHR: Low Energy High Resolution    (for Tc99m)
    Collimator MEGP: Medium Energy General Purpose (for In111, Lu177)
    Collimator HEGP: High Energy General Purpose   (for I131)

    """
    add_materials(sim)

    # check overlap
    # sim.check_volumes_overlap = False # set to True for debug

    # spect head
    head, lead_cover = add_spect_box(sim, name, crystal_size)

    # spect head
    crystal = add_crystal(sim, name, lead_cover, crystal_size)

    # spect collimator
    colli = None
    if collimator_type:
        colli = add_collimator(sim, name, head, collimator_type, rotation_deg, debug)

    # default orientation
    set_default_orientation(head, collimator_type)

    return head, colli, crystal


def set_default_orientation(head, collimator_type):
    # default head translation to set the shielding_front at position 0
    # (this translation is considered in rotate_gantry)
    p = get_geometrical_parameters()
    if collimator_type:
        tr = p[collimator_type].half_box_size - p[collimator_type].psd
    else:
        tr = 0
    head.translation = [0, tr, 0]

    # also the orientation is set such as Y is the normal of the detector
    rot = Rotation.from_euler("xy", (90, 0), degrees=True)
    head.rotation = rot.as_matrix()


def add_spect_two_heads(
    sim, name="spect", collimator_type="lehr", debug=False, radius=None
):
    cm = g4_units.cm
    if radius is None:
        radius = 36 * cm
    head1, colli1, crystal1 = add_spect_head(
        sim, f"{name}_1", collimator_type, debug=debug
    )

    # the second head is the same at 180 degrees
    head2, colli2, crystal2 = add_spect_head(
        sim, f"{name}_2", collimator_type, debug=debug
    )

    # set at their initial position
    set_default_orientation(head1, collimator_type)
    rotate_gantry(head1, radius, start_angle_deg=0, step_angle_deg=1, nb_angle=1)
    set_default_orientation(head2, collimator_type)
    rotate_gantry(head2, radius, start_angle_deg=180, step_angle_deg=1, nb_angle=1)

    return [head1, head2], [crystal1, crystal2]


def distance_to_center_of_crystal(sim, name="spect"):
    lead_cover = sim.volume_manager.volumes[f"{name}_lead_cover"]
    crystal = sim.volume_manager.volumes[f"{name}_crystal"]
    # distance from center to center of crystal
    shielding = sim.volume_manager.volumes[f"{name}_shielding"]
    d = shielding.translation[2] + lead_cover.translation[2] + crystal.translation[2]
    return d


def add_spect_box(sim, name, crystal_size):
    cm = g4_units.cm

    # colors
    blue = [0.5, 0.5, 1, 0.8]
    gray = [0.5, 0.5, 0.5, 1]
    white = [1, 1, 1, 1]
    yellow = [1, 1, 0, 1]
    green = [0, 1, 0, 1]

    # the total length
    spect_length = 19 * cm

    # bounding box
    head = sim.add_volume("Box", name)
    head.material = "G4_AIR"
    head.size = [57.6 * cm, 44.6 * cm, spect_length]
    head.color = white

    # shielding
    shielding = sim.add_volume("Box", f"{name}_shielding")
    shielding.mother = head.name
    shielding.size = head.size.copy()
    shielding.size[2] = 11.1375 * cm
    shielding.translation = [0, 0, -3.64 * cm]
    shielding.material = "Steel"
    shielding.color = yellow

    # shielding lead cover
    lead_cover = sim.add_volume("Box", f"{name}_lead_cover")
    lead_cover.mother = shielding.name
    lead_cover.size = [57.6 * cm, 40.6 * cm, 10.1375 * cm]
    lead_cover.translation = [0, 0, 0.5 * cm]
    lead_cover.material = "Lead"
    lead_cover.color = gray

    # crystal size
    if crystal_size == "3/8":
        cs = 0.9525 * cm
    else:
        cs = 1.5875 * cm

    # shielding alu cover
    alu_cover = sim.add_volume("Box", f"{name}_alu_cover")
    alu_cover.mother = lead_cover.name
    alu_cover.size = [54 * cm, 40 * cm, 0.13 * cm]
    alu_cover.translation = [0, 0, 5.00375 * cm]
    alu_cover.material = "Aluminium"
    alu_cover.color = blue

    # shielding reflector
    reflector = sim.add_volume("Box", f"{name}_reflector")
    reflector.mother = lead_cover.name
    reflector.size = [54 * cm, 40 * cm, 0.12 * cm]
    # reflector.translation = [0, 0, 3.92625 * cm] # for 3/8
    reflector.translation[2] = (
        lead_cover.size[2] / 2 - alu_cover.size[2] - cs - reflector.size[2] / 2
    )
    reflector.material = "TiO2"
    reflector.color = green

    # backside
    # The back-side is fairly complex, and may have a strong influence on the
    # spectrum: the model shown here is simplified
    backside = sim.add_volume("Box", f"{name}_backside")
    backside.mother = lead_cover.name
    h = alu_cover.size[2] + cs + reflector.size[2]
    backside.size = [54 * cm, 40 * cm, 8 * cm]
    # backside.translation = [0, 0, -0.13375 * cm] # for 3/8
    t = lead_cover.size[2] / 2 - h - backside.size[2] / 2
    backside.translation[2] = t
    backside.material = "Pyrex66"
    backside.color = blue

    return head, lead_cover


def add_crystal(sim, name, lead_cover, crystal_size):
    cm = g4_units.cm
    yellow = [1, 1, 0, 1]
    alu_cover_z = 0.13 * cm
    ref_z = 0.12 * cm
    # mono-bloc crystal thickness 3/8 of inch = 0.9525 cm
    # (if 5/8 inch = 1.5875 ; but probably need to translate elements)
    crystal = sim.add_volume("Box", f"{name}_crystal")
    crystal.mother = lead_cover.name
    if crystal_size == "3/8":
        crystal.size = [54 * cm, 40 * cm, 0.9525 * cm]
    else:
        crystal.size = [54 * cm, 40 * cm, 1.5875 * cm]
    crystal.translation[2] = lead_cover.size[2] / 2 - crystal.size[2] / 2 - alu_cover_z
    crystal.material = "NaITl"
    crystal.color = yellow
    return crystal


def add_collimator(sim, name, head, collimator_type, rotation_deg, debug):
    """
    Start with default lehr collimator description,
    then change some parameters for the other types
    """
    cm = g4_units.cm

    # colors
    red = [1, 0.7, 0.7, 0.8]
    blue = [0.5, 0.5, 1, 0.8]
    green = [0, 1, 0, 1]

    # mono-bloc crystal thickness 3/8 of inch
    colli_trd = sim.add_volume("Trd", f"{name}_collimator_trd")
    colli_trd.mother = head.name
    colli_trd.dx2 = 56.8 * cm / 2.0
    colli_trd.dy2 = 42.8 * cm / 2.0
    colli_trd.dx1 = 57.6 * cm / 2.0
    colli_trd.dy1 = 44.6 * cm / 2.0
    colli_trd.dz = 4.18 * cm / 2.0  # dep. on colli type
    colli_trd.translation = [0, 0, 4.02 * cm]  # dep. on colli type
    colli_trd.material = "G4_AIR"
    colli_trd.color = red

    # PSD (Position Sensitive Detection)
    psd = sim.add_volume("Box", f"{name}_collimator_psd")
    psd.mother = colli_trd.name
    psd.size = [54.6 * cm, 40.6 * cm, 0.1 * cm]
    psd.translation = [0, 0, 2.04 * cm]  # dep. on colli type
    psd.material = "Aluminium"
    psd.color = green

    # PSD layer
    psd_layer = sim.add_volume("Box", f"{name}_collimator_psd_layer")
    psd_layer.mother = colli_trd.name
    psd_layer.size = [54.6 * cm, 40.6 * cm, 0.15 * cm]
    psd_layer.translation = [0, 0, 1.915 * cm]  # dep. on colli type
    psd_layer.material = "PVC"
    psd_layer.color = red

    # Alu cover
    alu_cover = sim.add_volume("Box", f"{name}_collimator_alu_cover")
    alu_cover.mother = colli_trd.name
    alu_cover.size = [54.6 * cm, 40.6 * cm, 0.05 * cm]
    alu_cover.translation = [0, 0, -2.065 * cm]  # dep. on colli type
    alu_cover.material = "Aluminium"
    alu_cover.color = blue

    # air gap
    air_gap = sim.add_volume("Box", f"{name}_collimator_air_gap")
    air_gap.mother = colli_trd.name
    air_gap.size = [54.6 * cm, 40.6 * cm, 0.38 * cm]
    air_gap.translation = [0, 0, 1.65 * cm]  # dep. on colli type
    air_gap.material = "G4_AIR"
    air_gap.color = blue

    # core
    core = sim.add_volume("Box", f"{name}_collimator_core")
    core.mother = colli_trd.name
    core.size = [54.6 * cm, 40.6 * cm, 3.5 * cm]  # dep. on colli type
    core.translation = [0, 0, -0.29 * cm]
    core.material = "Lead"
    core.color = blue

    # adapt according to collimator type
    if collimator_type == "megp":
        colli_trd.dz = 6.48 * cm / 2.0
        colli_trd.translation = [0, 0, 5.17 * cm]
        psd.translation = [0, 0, 3.19 * cm]
        psd_layer.translation = [0, 0, 3.065 * cm]
        alu_cover.translation = [0, 0, -3.215 * cm]
        air_gap.translation = [0, 0, 2.80 * cm]
        core.size = [54.6 * cm, 40.6 * cm, 5.8 * cm]

    if collimator_type == "hegp":
        colli_trd.dz = 7.28 * cm / 2.0
        colli_trd.translation = [0, 0, 5.57 * cm]
        psd.translation = [0, 0, 3.59 * cm]
        psd_layer.translation = [0, 0, 3.465 * cm]
        alu_cover.translation = [0, 0, -3.615 * cm]
        air_gap.translation = [0, 0, 3.2 * cm]
        core.size = [54.6 * cm, 40.6 * cm, 6.6 * cm]

    # repeater for the holes
    holep = None
    if collimator_type == "megp":
        holep = megp_collimator_repeater(sim, name, core, debug)
    if collimator_type == "lehr":
        holep = lehr_collimator_repeater(sim, name, core, debug)
    if collimator_type == "hegp":
        holep = hegp_collimator_repeater(sim, name, core, debug)
    if not holep:
        fatal(
            f"Error, unknown collimator type {collimator_type}. "
            f'Use "megp" or "lehr" or "hegp" or "False"'
        )

    # should we rotate the collimator holes ?
    holep.rotation = Rotation.from_euler("Z", rotation_deg, degrees=True).as_matrix()

    return colli_trd


def hegp_collimator_repeater(sim, name, core, debug):
    cm = g4_units.cm
    mm = g4_units.mm
    # one single hole
    hole = sim.add_volume("Hexagon", f"{name}_collimator_hole")
    hole.height = 6.6 * cm
    hole.radius = 0.2 * cm
    hole.material = "G4_AIR"
    hole.mother = core.name

    # parameterised holes
    holep = RepeatParametrisedVolume(repeated_volume=hole)
    holep.linear_repeat = [54, 70, 1]
    if debug:
        holep.linear_repeat = [10, 10, 1]
    holep.translation = [10.0459 * mm, 5.8 * mm, 0]
    # dot it twice, with the following offset
    holep.offset_nb = 2
    holep.offset = [5.0229 * mm, 2.9000 * mm, 0]

    sim.volume_manager.add_volume(holep)
    return holep


def megp_collimator_repeater(sim, name, core, debug):
    cm = g4_units.cm
    mm = g4_units.mm
    # one single hole
    hole = sim.add_volume("Hexagon", f"{name}_collimator_hole")
    hole.height = 5.8 * cm
    hole.radius = 0.15 * cm
    hole.material = "G4_AIR"
    hole.mother = core.name

    # parameterised holes
    holep = RepeatParametrisedVolume(repeated_volume=hole)
    if debug:
        holep.linear_repeat = [10, 10, 1]
    else:
        holep.linear_repeat = [77, 100, 1]

    holep.translation = [7.01481 * mm, 4.05 * mm, 0]
    # do it twice, with the following offset
    holep.offset_nb = 2
    holep.offset = [3.50704 * mm, 2.025 * mm, 0]

    sim.volume_manager.add_volume(holep)
    return holep


def lehr_collimator_repeater(sim, name, core, debug):
    cm = g4_units.cm
    mm = g4_units.mm
    # one single hole
    hole = HexagonVolume(name=f"{name}_collimator_hole")
    hole.height = 3.5 * cm
    hole.radius = 0.075 * cm
    hole.material = "G4_AIR"
    hole.mother = core.name
    hole.build_physical_volume = False
    sim.volume_manager.add_volume(hole)

    # parameterised holes
    holep = RepeatParametrisedVolume(repeated_volume=hole)
    if debug:
        holep.linear_repeat = [30, 30, 1]
    else:
        holep.linear_repeat = [183, 235, 1]

    holep.translation = [2.94449 * mm, 1.7 * mm, 0]

    # do it twice, with the following offset
    holep.offset_nb = 2
    holep.offset = [1.47224 * mm, 0.85 * mm, 0]

    sim.volume_manager.add_volume(holep)
    return holep


def lehr_collimator_repeater2_WIP(sim, name, core, debug):
    cm = g4_units.cm
    mm = g4_units.mm

    # one hole
    hole_a = HexagonVolume(name=f"{name}_collimator_hole_a")
    hole_a.height = 3.5 * cm
    hole_a.radius = 0.075 * cm
    hole_a.material = "G4_AIR"
    hole_a.mother = core.name

    # double hole
    hole = unite_volumes(hole_a, hole_a, translation=[1.47224 * mm, 0.85 * mm, 0])
    sim.volume_manager.add_volume(hole)

    # parameterised holes
    holep = RepeatParametrisedVolume(repeated_volume=hole)
    if debug:
        holep.linear_repeat = [30, 30, 1]
    else:
        holep.linear_repeat = [183, 235, 1]

    holep.translation = [2.94449 * mm, 1.7 * mm, 0]
    sim.volume_manager.add_volume(holep)

    return holep


def lehr_collimator_repeater_noparam_WIP(sim, name, core, debug):
    cm = g4_units.cm
    mm = g4_units.mm
    # one single hole
    hole = sim.add_volume("Hexagon", name=f"{name}_collimator_hole")
    hole.height = 3.5 * cm
    hole.radius = 0.075 * cm
    hole.material = "G4_AIR"
    hole.mother = core.name
    translations = []

    if debug:
        linear_repeat = [30, 30, 1]
    else:
        linear_repeat = [183, 235, 1]

    offset_nb = 2
    translation = [2.94449 * mm, 1.7 * mm, 0]
    offset = [1.47224 * mm, 0.85 * mm, 0]
    start = [-(x - 1) * y / 2.0 for x, y in zip(linear_repeat, translation)]
    for i in range(linear_repeat[0]):
        for j in range(linear_repeat[1]):
            for k in range(linear_repeat[2]):
                for o in range(offset_nb):
                    tr_x = start[0] + i * translation[0] + o * offset[0]
                    tr_y = start[1] + j * translation[1] + o * offset[1]
                    tr_z = start[2] + k * translation[2] + o * offset[2]
                    tr = [tr_x, tr_y, tr_z]
                    translations.append(tr)
    hole.translation = translations  # repeated

    return hole


def add_simplified_digitizer_tc99m_OLD(
    sim, crystal_volume_name, output_name, scatter_flag=False
):
    # units
    keV = g4_units.keV
    # default  channels
    channels = []
    if scatter_flag:
        channels = [
            {
                "name": f"scatter_{crystal_volume_name}",
                "min": 114 * keV,
                "max": 126 * keV,
            }
        ]
    channels.append(
        {
            "name": f"peak140_{crystal_volume_name}",
            "min": 126 * keV,
            "max": 154.55 * keV,
        }
    )
    proj = add_digitizer_OLD(sim, crystal_volume_name, channels)
    # output
    proj.output_filename = output_name
    return proj


def add_digitizer_OLD(sim, crystal_volume_name, channels):
    # units
    mm = g4_units.mm
    cc = add_digitizer_energy_windows_OLD(sim, crystal_volume_name, channels)

    # projection
    proj = sim.add_actor(
        "DigitizerProjectionActor", f"Projection_{crystal_volume_name}"
    )
    proj.attached_to = cc.attached_to
    proj.input_digi_collections = [x["name"] for x in cc.channels]
    # proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.spacing = [5 * mm, 5 * mm]
    proj.size = [128, 128]

    return proj


def add_digitizer_energy_windows_OLD(sim, crystal_volume_name, channels):
    # Hits
    hc = sim.add_actor("DigitizerHitsCollectionActor", f"Hits_{crystal_volume_name}")
    hc.attached_to = crystal_volume_name
    # no output by default
    hc.output_filename = None
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # Singles
    sc = sim.add_actor("DigitizerAdderActor", f"Singles_{crystal_volume_name}")
    sc.attached_to = hc.attached_to
    sc.input_digi_collection = hc.name
    sc.policy = "EnergyWinnerPosition"
    sc.output_filename = None

    # energy windows
    cc = sim.add_actor(
        "DigitizerEnergyWindowsActor", f"EnergyWindows_{crystal_volume_name}"
    )
    cc.attached_to = sc.attached_to
    cc.input_digi_collection = sc.name
    cc.channels = channels
    cc.output_filename = None

    return cc


def add_digitizer_tc99m_OLD(sim, crystal_name, name, spectrum_channel=True):
    # create main chain
    mm = g4_units.mm
    digitizer = Digitizer(sim, crystal_name, name)

    # Singles
    sc = digitizer.add_module("DigitizerAdderActor", f"{name}_singles")
    sc.group_volume = None
    sc.policy = "EnergyWinnerPosition"

    # detection efficiency
    ea = digitizer.add_module("DigitizerEfficiencyActor", f"{name}_eff")
    ea.efficiency = 0.86481  # FAKE

    # energy blurring
    keV = g4_units.keV
    eb = digitizer.add_module("DigitizerBlurringActor", f"{name}_blur")
    eb.blur_attribute = "TotalEnergyDeposit"
    eb.blur_method = "InverseSquare"
    eb.blur_resolution = 0.063  # FAKE
    eb.blur_reference_value = 140.57 * keV

    # spatial blurring
    sb = digitizer.add_module("DigitizerSpatialBlurringActor", f"{name}_sp_blur")
    sb.blur_attribute = "PostPosition"
    sb.blur_fwhm = 7.6 * mm  # FAKE
    sb.keep_in_solid_limits = True
    sb.use_truncated_Gaussian = False

    # energy windows (Energy range. 35-588 keV)
    cc = digitizer.add_module("DigitizerEnergyWindowsActor", f"{name}_energy_window")
    channels = [
        {"name": f"spectrum", "min": 3 * keV, "max": 160 * keV},
        {"name": f"scatter", "min": 108.57749938965 * keV, "max": 129.5924987793 * keV},
        {"name": f"peak140", "min": 129.5924987793 * keV, "max": 150.60751342773 * keV},
    ]
    if not spectrum_channel:
        channels.pop(0)
    cc.channels = channels

    # projection
    proj = digitizer.add_module("DigitizerProjectionActor", f"{name}_projection")
    channel_names = [c["name"] for c in channels]
    proj.input_digi_collections = channel_names
    proj.spacing = [2.21 * mm * 2, 2.21 * mm * 2]
    proj.size = [128, 128]

    # end
    return digitizer


def add_digitizer_tc99m_v2_OLD(sim, crystal_name, name, spectrum_channel=True):
    # create the main chain
    mm = g4_units.mm
    digitizer = Digitizer(sim, crystal_name, name)

    # Singles
    sc = digitizer.add_module("DigitizerAdderActor", f"{name}_singles")
    sc.group_volume = None
    sc.policy = "EnergyWinnerPosition"

    # detection efficiency
    # ea = digitizer.add_module("DigitizerEfficiencyActor", f"{name}_eff")
    # ea.efficiency = 0.86481 # FAKE

    # energy blurring
    keV = g4_units.keV
    eb = digitizer.add_module("DigitizerBlurringActor", f"{name}_blur")
    eb.blur_attribute = "TotalEnergyDeposit"
    eb.blur_method = "InverseSquare"
    eb.blur_resolution = 0.063  # FAKE
    eb.blur_reference_value = 140.57 * keV

    # spatial blurring
    sb = digitizer.add_module("DigitizerSpatialBlurringActor", f"{name}_sp_blur")
    sb.blur_attribute = "PostPosition"
    sb.blur_fwhm = 3 * mm  # FAKE
    sb.keep_in_solid_limits = True
    sb.use_truncated_Gaussian = False

    # energy windows (Energy range. 35-588 keV)
    cc = digitizer.add_module("DigitizerEnergyWindowsActor", f"{name}_energy_window")
    channels = [
        {"name": f"spectrum", "min": 3 * keV, "max": 160 * keV},
        {"name": f"scatter", "min": 108.57749938965 * keV, "max": 129.5924987793 * keV},
        {"name": f"peak140", "min": 129.5924987793 * keV, "max": 150.60751342773 * keV},
    ]
    if not spectrum_channel:
        channels.pop(0)
    cc.channels = channels

    # projection
    proj = digitizer.add_module("DigitizerProjectionActor", f"{name}_projection")
    channel_names = [c["name"] for c in channels]
    proj.input_digi_collections = channel_names
    proj.spacing = [2.21 * mm * 2, 2.21 * mm * 2]
    proj.size = [128, 128]
    proj.write_to_disk = True

    # here, we need this rotation
    proj.detector_orientation_matrix = Rotation.from_euler(
        "yx", (0, 180), degrees=True  # 0 180 => like intevo
    ).as_matrix()

    # end
    return digitizer


def add_digitizer_lu177_OLD(sim, crystal_name, name, spectrum_channel=True):
    # create main chain
    mm = g4_units.mm
    digitizer = Digitizer(sim, crystal_name, name)

    # Singles
    sc = digitizer.add_module("DigitizerAdderActor", f"{name}_singles")
    sc.group_volume = None
    sc.policy = "EnergyWinnerPosition"

    # detection efficiency
    ea = digitizer.add_module("DigitizerEfficiencyActor")
    ea.efficiency = 0.86481  # FAKE

    # energy blurring
    keV = g4_units.keV
    eb = digitizer.add_module("DigitizerBlurringActor")
    eb.blur_attribute = "TotalEnergyDeposit"
    eb.blur_method = "InverseSquare"
    eb.blur_resolution = 0.063  # FAKE
    eb.blur_reference_value = 140.57 * keV
    # eb.blur_resolution = 0.13
    # eb.blur_reference_value = 80 * keV
    # eb.blur_slope = -0.09 * 1 / MeV  # fixme unsure about unit

    # spatial blurring
    # Source: HE4SPECS - FWHM = 3.9 mm
    # FWHM = 2.sigma.sqrt(2ln2) -> sigma = 1.656 mm
    sb = digitizer.add_module("DigitizerSpatialBlurringActor")
    sb.blur_attribute = "PostPosition"
    sb.blur_fwhm = 7.6 * mm  # FAKE
    sb.keep_in_solid_limits = True
    sb.use_truncated_Gaussian = False

    # energy windows (Energy range. 35-588 keV)
    cc = digitizer.add_module("DigitizerEnergyWindowsActor", f"{name}_energy_window")
    keV = g4_units.keV
    # 112.9498 keV  = 6.20 %
    # 208.3662 keV  = 10.38 %
    p1 = 112.9498 * keV
    p2 = 208.3662 * keV
    channels = [
        {"name": "spectrum", "min": 35 * keV, "max": 588 * keV},
        *energy_windows_peak_scatter("peak113", "scatter1", "scatter2", p1, 0.2, 0.1),
        *energy_windows_peak_scatter("peak208", "scatter3", "scatter4", p2, 0.2, 0.1),
    ]
    if not spectrum_channel:
        channels.pop(0)
    cc.channels = channels

    # projection
    proj = digitizer.add_module("DigitizerProjectionActor", f"{name}_projection")
    channel_names = [c["name"] for c in channels]
    proj.input_digi_collections = channel_names
    proj.spacing = [2.21 * mm * 2, 2.21 * mm * 2]
    proj.size = [128, 128]
    proj.write_to_disk = True

    # end
    return digitizer


def add_digitizer_lu177_v2_OLD(sim, crystal_name, name, spectrum_channel=True):
    # create main chain
    mm = g4_units.mm
    digitizer = Digitizer(sim, crystal_name, name)

    # Singles
    sc = digitizer.add_module("DigitizerAdderActor", f"{name}_singles")
    sc.group_volume = None
    sc.policy = "EnergyWinnerPosition"

    # energy blurring
    keV = g4_units.keV
    eb = digitizer.add_module("DigitizerBlurringActor")
    eb.blur_attribute = "TotalEnergyDeposit"
    eb.blur_method = "InverseSquare"
    eb.blur_resolution = 0.063  # FAKE
    eb.blur_reference_value = 140.57 * keV

    # spatial blurring
    # Source: HE4SPECS - FWHM = 3.9 mm
    # FWHM = 2.sigma.sqrt(2ln2) -> sigma = 1.656 mm
    sb = digitizer.add_module("DigitizerSpatialBlurringActor")
    sb.blur_attribute = "PostPosition"
    sb.blur_fwhm = 3.9 * mm  # FAKE
    sb.keep_in_solid_limits = True
    sb.use_truncated_Gaussian = False

    # energy windows (Energy range. 35-588 keV)
    cc = digitizer.add_module("DigitizerEnergyWindowsActor", f"{name}_energy_window")
    keV = g4_units.keV
    # 112.9498 keV  = 6.20 %
    # 208.3662 keV  = 10.38 %
    p1 = 112.9498 * keV
    p2 = 208.3662 * keV
    channels = [
        {"name": "spectrum", "min": 35 * keV, "max": 588 * keV},
        *energy_windows_peak_scatter("peak113", "scatter1", "scatter2", p1, 0.2, 0.1),
        *energy_windows_peak_scatter("peak208", "scatter3", "scatter4", p2, 0.2, 0.1),
    ]
    if not spectrum_channel:
        channels.pop(0)
    cc.channels = channels

    # projection
    proj = digitizer.add_module("DigitizerProjectionActor", f"{name}_projection")
    channel_names = [c["name"] for c in channels]
    proj.input_digi_collections = channel_names
    proj.spacing = [2.21 * mm * 2, 2.21 * mm * 2]
    proj.size = [128, 128]
    proj.write_to_disk = True

    # end
    return digitizer


def add_digitizer_lu177_v3_OLD(sim, crystal_name, name, spectrum_channel=True):
    digitizer = add_digitizer_lu177_v2_OLD(sim, crystal_name, name, spectrum_channel)

    # here, we need this rotation
    proj = digitizer.find_module_by_type("DigitizerProjectionActor")
    proj.detector_orientation_matrix = Rotation.from_euler(
        "yx", (0, 180), degrees=True  # 0 180 => like intevo
    ).as_matrix()

    # end
    return digitizer


def compute_plane_position_and_distance_to_crystal(collimator_type):
    sim = Simulation()
    spect, colli, crystal = add_spect_head(sim, "spect", collimator_type, debug=True)
    pos = get_volume_position_in_head(sim, "spect", "collimator_psd", "max", axis=2)
    y = get_volume_position_in_head(sim, "spect", "crystal", "min", axis=2)
    crystal_distance = pos - y
    psd = spect.size[2] / 2.0 - pos
    return pos, crystal_distance, psd


def get_plane_position_and_distance_to_crystal_OLD(collimator_type):
    """
    This has been computed with t043_distances or compute_plane_position_and_distance_to_crystal
    - first : distance from head center to the PSD (translation for the plane)
    - second: distance from PSD to center of the crystal
    - third : distance from the head boundary to the PSD (for spect_radius info)
    """
    if collimator_type == "lehr":
        return 61.1, 47.875, 33.9

    if collimator_type == "megp":
        return 84.1, 70.875, 10.9

    if collimator_type == "hegp":
        return 92.1, 78.875, 2.9

    fatal(
        f'Unknown collimator type "{collimator_type}", please use lehr or megp or hegp'
    )


def set_head_orientation(head, collimator_type, radius, gantry_angle_deg=0):
    # pos is the distance from entrance detection plane and head boundary
    pos, _, _ = compute_plane_position_and_distance_to_crystal(collimator_type)
    distance = radius + pos
    # rotation X180 is to set the detector head-foot
    # rotation Z90 is the gantry angle
    r1 = Rotation.from_euler("X", 90, degrees=True)
    r2 = Rotation.from_euler("Z", gantry_angle_deg, degrees=True)
    r = r2 * r1
    head.translation = r.apply([0, 0, -distance])
    head.rotation = r.as_matrix()
    return r


def add_detection_plane_for_arf_OLD(
    sim, plane_size, colli_type, radius, gantry_angle_deg=0, det_name=None
):
    if det_name is None:
        det_name = "arf_plane"

    # rotation ? no
    r = Rotation.from_euler("yx", (0, 0), degrees=True)

    # plane
    nm = g4_units.nm
    detector_plane = sim.add_volume("Box", det_name)
    detector_plane.material = "G4_Galactic"
    detector_plane.color = [1, 0, 0, 1]
    detector_plane.size = [plane_size[0], plane_size[1], 1 * nm]

    # (fake) initial head rotation
    head = Box()
    head.translation = None
    head.rotation = None
    ri = set_head_orientation(head, colli_type, radius, gantry_angle_deg)

    # orientation
    detector_plane.rotation = (ri * r).as_matrix()
    detector_plane.translation = ri.apply([0, 0, -radius])

    return detector_plane


def get_normal_to_arf_plane():
    return [0, 0, 1]


def get_normal_to_detector():
    return [0, 0, -1]


def add_detection_plane_for_arf(sim, det_name, colli_type, plane_size=None):
    # the plane is in the world coordinate system outside the real spect head box.

    # user plane size only for debug purpose
    mm = g4_units.mm
    if plane_size is None:
        plane_size = [576 * mm, 446 * mm]

    # plane
    nm = g4_units.nm
    detector_plane = sim.add_volume("Box", det_name)
    detector_plane.material = "G4_Galactic"
    detector_plane.color = [1, 0, 0, 1]
    detector_plane.size = [plane_size[0], plane_size[1], 1 * nm]

    # compute the position according to the front psd
    p = get_geometrical_parameters()
    arf_position = p[colli_type].psd
    detector_plane.translation = [0, -arf_position, 0]

    # initial_rotation like the detector
    r = Rotation.from_euler("x", -90, degrees=True)
    detector_plane.rotation = r.as_matrix()

    # DEBUG
    """fake = sim.add_volume("Box", "fake")
    fake.mother = detector_plane
    fake.material = "G4_Galactic"
    fake.color = [1, 1, 0, 1]
    fake.size = [20 * mm, 10 * mm, 2 * mm]
    fake.translation = [-plane_size[0] / 2, -plane_size[1] / 2, 0]"""

    # rotate
    rotate_gantry(detector_plane, radius=0, start_angle_deg=0)
    return detector_plane


def add_source_for_arf_training_dataset(
    sim, source_name, activity, detector_plane, min_energy, max_energy
):
    cm = g4_units.cm
    source = sim.add_source("GenericSource", source_name)
    source.particle = "gamma"
    source.activity = activity
    source.position.type = "disc"
    source.position.radius = 5 * cm
    source.position.translation = [0, -20 * cm, 0]
    source.direction.type = "iso"
    source.energy.type = "range"
    source.energy.min_energy = min_energy
    source.energy.max_energy = max_energy
    source.direction.acceptance_angle.volumes = [detector_plane.name]
    source.direction.acceptance_angle.intersection_flag = True

    return source


def add_actor_for_arf_training_dataset(sim, colli_type, ene_win_actor, rr):
    # the detector is in front of the spect head volume, outside
    detector_plane = add_detection_plane_for_arf(sim, "arf_plane", colli_type)

    # arf actor for building the training dataset
    arf = sim.add_actor("ARFTrainingDatasetActor", "ARF (training)")
    arf.energy_windows_actor = ene_win_actor.name
    arf.attached_to = detector_plane.name
    arf.output_filename = "arf_training_dataset.root"
    arf.russian_roulette = rr
    arf.plane_axis = [0, 1, 2]  # the depth is plane_axis[2] = 2

    return detector_plane, arf


def add_arf_detector(sim, name, colli_type, image_size, image_spacing, pth_filename):
    # plane_size = [image_size[0] * image_spacing[0], image_size[1] * image_spacing[1]]
    # print(f"plane_size: {plane_size}")
    det_plane = add_detection_plane_for_arf(sim, det_name=name, colli_type=colli_type)

    # set the position in front of the collimator
    p = get_geometrical_parameters()
    crystal_distance = p[colli_type].crystal_distance

    arf = sim.add_actor("ARFActor", f"{name}_arf")
    arf.attached_to = det_plane.name
    arf.output_filename = f"projection_{arf.name}.mhd"
    arf.batch_size = 1e5
    arf.image_size = image_size
    arf.image_spacing = image_spacing
    arf.verbose_batch = False
    print(f"crystal_distance: {crystal_distance}")
    arf.distance_to_crystal = crystal_distance
    arf.pth_filename = pth_filename
    arf.flip_plane = False
    arf.plane_axis = [0, 1, 2]
    arf.enable_hit_slice = False
    arf.gpu_mode = "auto"

    return det_plane, arf


def rotate_gantry(head, radius, start_angle_deg, step_angle_deg=1, nb_angle=1):
    # compute the nb translation and rotation
    # WARNING: use the initial rotation and translation of the head.
    translations = []
    rotations = []
    current_angle_deg = start_angle_deg
    for r in range(nb_angle):
        tr = head.translation.copy()
        initial_rotation = head.rotation
        tr[1] += radius
        t, rot = get_transform_orbiting(tr, "Z", current_angle_deg)
        rot = Rotation.from_matrix(rot)
        rot = rot.as_matrix()
        rot = np.matmul(rot, initial_rotation)
        translations.append(t)
        rotations.append(rot)
        current_angle_deg += step_angle_deg

    # set the motion for the SPECT head
    if nb_angle > 1:
        head.add_dynamic_parametrisation(translation=translations, rotation=rotations)
    # we set the initial position in all cases, this allows for check_overlap to be done
    # with the first position
    head.translation = translations[0]
    head.rotation = rotations[0]


def add_actor_for_arf_training_dataset_OLD(sim, head, colli_type, ene_win_actor, rr):
    nm = g4_units.nm
    cm = g4_units.cm

    # detector input plane
    pos, crystal_dist, psd = get_plane_position_and_distance_to_crystal_OLD(colli_type)
    pos += 1 * nm  # to avoid overlap
    detector_plane = sim.add_volume("Box", "arf_plane")
    detector_plane.mother = head
    detector_plane.size = [57.6 * cm, 44.6 * cm, 1 * nm]
    detector_plane.translation = [0, 0, pos]
    detector_plane.material = "G4_Galactic"
    detector_plane.color = [1, 0, 0, 1]

    # arf actor for building the training dataset
    arf = sim.add_actor("ARFTrainingDatasetActor", "ARF (training)")
    arf.attached_to = detector_plane.name
    arf.output_filename = "arf_training_dataset.root"
    arf.energy_windows_actor = ene_win_actor.name
    arf.russian_roulette = rr

    return detector_plane, arf


def add_arf_detector_OLD(
    sim, radius, gantry_angle, size, spacing, colli_type, name, i, pth
):
    plane_size = [size[0] * spacing[0], size[1] * spacing[1]]
    det_plane = add_detection_plane_for_arf_OLD(
        sim,
        plane_size,
        colli_type=colli_type,
        radius=radius,
        gantry_angle_deg=gantry_angle,
        det_name=f"{name}_{i}",
    )

    pos, crystal_distance, psd = compute_plane_position_and_distance_to_crystal(
        colli_type
    )

    arf = sim.add_actor("ARFActor", f"{name}_arf_{i}")
    arf.attached_to = det_plane.name
    arf.output_filename = f"projection_{i}.mhd"
    arf.batch_size = 1e5
    arf.image_size = size
    arf.image_spacing = spacing
    arf.verbose_batch = False
    arf.distance_to_crystal = crystal_distance  # 74.625 * mm
    arf.pth_filename = pth
    arf.flip_plane = True  # because the training was backside
    arf.enable_hit_slice = False
    arf.gpu_mode = "auto"

    return det_plane, arf


def get_pytomography_detector_physics_data(colli_name):
    cm = g4_units.cm
    # create a fake simulation to get the volume information
    sim = Simulation()
    det, colli, crystal = add_spect_head(sim, f"fake", collimator_type=colli_name)
    holep = sim.volume_manager.find_volumes("collimator_hole_param")[0]
    hole = sim.volume_manager.find_volumes("collimator_hole")[0]
    d = {
        "hole_shape": 6,
        "hole_diameter": hole.radius * 2 / cm,
        "hole_spacing": holep.translation[1] / cm,
        "collimator_thickness": hole.height / cm,
        "collimator_material": colli.material.lower(),
        "crystal_width": crystal.size[1] / cm,
        "crystal_height": crystal.size[2] / cm,
    }

    return d


def get_default_size_and_spacing():
    mm = g4_units.mm
    return [128, 128], [2.21 * mm * 2, 2.21 * mm * 2]


def add_digitizer(
    sim, crystal_name, name=None, size=None, spacing=None, channels=None, filename=None
):
    # default parameters
    mm = g4_units.mm
    if name is None:
        name = crystal_name
    if size is None:
        size = get_default_size_and_spacing()[0]
    if spacing is None:
        spacing = get_default_size_and_spacing()[1]
    if channels is None:
        channels = get_default_energy_windows("tc99m")

    # create the main digitizer chain
    digitizer = Digitizer(sim, crystal_name, name)

    # Singles
    sc = digitizer.add_module("DigitizerAdderActor", f"{name}_singles")
    sc.group_volume = None
    sc.policy = "EnergyWinnerPosition"

    # detection efficiency
    # ea = digitizer.add_module("DigitizerEfficiencyActor")
    # ea.efficiency = 0.86481  # FAKE

    # energy blurring
    keV = g4_units.keV
    eb = digitizer.add_module("DigitizerBlurringActor", f"{name}_eblur")
    eb.blur_attribute = "TotalEnergyDeposit"
    eb.blur_method = "InverseSquare"
    eb.blur_resolution = 0.063  # in %, FAKE !
    eb.blur_reference_value = 140.57 * keV

    # spatial blurring
    sb = digitizer.add_module("DigitizerSpatialBlurringActor", f"{name}_sblur")
    sb.blur_attribute = "PostPosition"
    sb.blur_fwhm = 3 * mm  # FAKE !
    sb.keep_in_solid_limits = True
    sb.use_truncated_Gaussian = False

    # default energy windows (Energy range. 35-588 keV)
    cc = digitizer.add_module("DigitizerEnergyWindowsActor", f"{name}_energy_window")
    cc.channels = channels

    # projection
    proj = digitizer.add_module("DigitizerProjectionActor", f"{name}_projection")
    channel_names = [c["name"] for c in channels]
    proj.input_digi_collections = channel_names
    proj.spacing = spacing
    proj.size = size
    proj.write_to_disk = True
    if filename is not None:
        proj.output_filename = filename

    # projection plane: it depends on how the spect device is described
    # here, we need this rotation
    proj.detector_orientation_matrix = Rotation.from_euler(
        "yx", (0, 180), degrees=True  # 0 180 => like intevo
    ).as_matrix()

    return digitizer


def calculate_collimator_acceptance_angle(collimator_type, energy, prob_threshold):
    p = get_geometrical_parameters()
    hole_diameter = p[collimator_type].hole_diameter
    collimator_length = p[collimator_type].collimator_length
    septa_thickness = p[collimator_type].septa_thickness

    mu_lead_cm = get_mu_from_xraylib("Pb", energy)
    print(
        f"collimator_type: {collimator_type}, energy: {energy} keV, mu_lead_cm: {mu_lead_cm} cm-1"
    )
    print(f"collimator septa thicnkess {septa_thickness} mm")
    # theta_max = calculate_max_penetration_angle(
    #    hole_diameter, collimator_length, septa_thickness, mu_lead_cm, prob_threshold
    # )
    l_eff, theta_acc = calculate_acceptance_angle(
        hole_diameter, collimator_length, mu_lead_cm
    )
    print(f"theta_max: {theta_acc} deg")

    return l_eff, theta_acc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.linacs.elektaversa as versa
import opengate.contrib.linacs.dicomrtplan as rtplan
from opengate.tests import utility
from opengate.utility import g4_units
import numpy as np
import uproot


def add_volume_to_irradiate(sim, name):
    m = g4_units.m
    patient = sim.add_volume("Box", "water_plane")
    patient.mother = name
    patient.material = "G4_Galactic"
    patient.size = [1 * m, 0.5 * m, 0.25 * m]
    patient.color = [1, 0, 0, 1]  # red
    return patient


def add_phase_space_isocenter(sim, name, pos):
    isocenter_sphere = sim.add_volume("Sphere", "sphere")
    isocenter_sphere.material = "G4_Galactic"
    isocenter_sphere.mother = name
    isocenter_sphere.translation = pos
    isocenter_sphere.rmax = 0
    isocenter_sphere.rmax = 1 * cm
    isocenter_sphere.color = [0, 1, 0, 1]  # red

    phsp = sim.add_actor("PhaseSpaceActor", f"phsp")
    phsp.attached_to = isocenter_sphere.name
    phsp.attributes = ["EventID"]
    return phsp


def add_alpha_source(sim, name, pos_Z, nb_part):
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm
    plan_source = sim.add_volume("Box", "plan_alpha_source")
    plan_source.material = "G4_Galactic"
    plan_source.mother = name
    plan_size = np.array([1 * nm, 1 * nm, 1 * nm])
    plan_source.size = np.copy(plan_size)
    plan_source.translation = [0 * mm, 0 * mm, -pos_Z / 2 + 300 * mm]

    source = sim.add_source("GenericSource", "alpha_source")
    MeV = gate.g4_units.MeV
    source.particle = "alpha"
    source.attached_to = plan_source.name
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV
    source.position.type = "box"
    source.position.size = np.copy(plan_size)
    source.direction.type = "momentum"
    source.direction_relative_to_attached_volume = True
    source.direction.momentum = [0, 0, -1]
    source.n = nb_part


def validation_test_19_rt_plan(array, nb_part):
    if len(array["EventID"]) == nb_part:
        return True
    else:
        return False


if __name__ == "__main__":
    # paths
    paths = utility.get_default_test_paths(__file__, output_folder="test019_linac")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    sim.number_of_threads = 1
    sim.output_dir = paths.output
    sim.random_seed = 123456789
    sim.check_volumes_overlap = True
    sim.output_dir = paths.output

    # unit
    nm = gate.g4_units.nm
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    sec = gate.g4_units.s
    Bq = gate.g4_units.Bq

    # world
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_Galactic"

    # materials and empty linac
    versa.add_linac_materials(sim)
    sad = 1000 * mm
    linac = versa.add_empty_linac_box(sim, "linac_box", sad)
    linac.material = "G4_Galactic"

    # add alpha source :
    nb_part = np.random.randint(1, 100)
    z_linac = linac.size[2]
    add_alpha_source(sim, linac.name, z_linac / 2 - 5.6 * mm, nb_part)

    # Linac head rotation and jaws and mlc translation according to a DICOM rt plan
    cp = np.random.randint(1, 230, 1)[0]
    print(cp)
    rt_plan_parameters = rtplan.read(str(paths.data / "DICOM_RT_plan.dcm"))
    versa.rotation_around_user_point(
        sim, linac.name, "x", rt_plan_parameters["gantry angle"][cp]
    )

    # The patient image is simulated at a position which is the center of the simulation
    # regardless the offset of the image
    # And the isocenter is given the real center of the image. That's why we have
    # to correct the isocenter of the real image center
    center_patient_on_DICOM_img = np.array([11 * mm, -194 * mm, 45 * mm])
    vector_translation_isocenter_to_center = (
        center_patient_on_DICOM_img - rt_plan_parameters["isocenter"][cp]
    )
    print(vector_translation_isocenter_to_center)

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 1000 * m)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # add water slice with a dose actor and a motion actor
    volume = add_volume_to_irradiate(sim, world.name)
    phsp = add_phase_space_isocenter(
        sim, volume.name, vector_translation_isocenter_to_center
    )
    phsp.output_filename = "phsp_versa_mlc_RT_plan.root"
    print(phsp.get_output_path())

    translation_volume = volume.translation
    new_translation = (
        np.array(translation_volume) - vector_translation_isocenter_to_center
    )
    volume.translation = new_translation

    # start simulation
    # The number of particles provided (sim.activity) will be adapted regarding
    # the number of MU delivered at each control points.
    sim.run()

    # print results
    print(stats)

    f_phsp = uproot.open(phsp.get_output_path())
    arr = f_phsp["phsp"].arrays()

    # test
    is_ok = validation_test_19_rt_plan(arr, nb_part)
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import uproot
from opengate.actors.coincidences import coincidences_sorter
from opengate.exception import GateImplementationError
import opengate.tests.utility as utility


g_cm3 = gate.g4_units.g_cm3
mm = gate.g4_units.mm
Bq = gate.g4_units.Bq
keV = gate.g4_units.keV
sec = gate.g4_units.second
ns = gate.g4_units.nanosecond
deg = gate.g4_units.degree


def create_materials(sim):
    sim.volume_manager.material_database.add_material_nb_atoms(
        "LYSO", ["Lu", "Y", "Si", "O"], [18, 2, 10, 50], 7.1 * g_cm3
    )


if __name__ == "__main__":

    is_ok = True
    exceptions = []
    paths = utility.get_default_test_paths(__file__, None, output_folder="test086")

    for adder_policy in ["EnergyWinnerPosition", "EnergyWeightedCentroidPosition"]:

        sim = gate.Simulation()
        sim.g4_verbose = False
        # sim.visu = False
        sim.visu_type = "vrml"
        sim.random_engine = "MersenneTwister"
        sim.random_seed = 1234
        sim.number_of_threads = 1
        sim.output_dir = paths.output
        create_materials(sim)

        sim.world.size = [200 * mm, 200 * mm, 200 * mm]
        sim.world.material = "G4_AIR"

        crystal = sim.add_volume("Box", "crystal")
        crystal.size = [50 * mm, 50 * mm, 50 * mm]
        crystal.translation = gate.geometry.utility.get_grid_repetition(
            [2, 1, 1], [60 * mm, 0, 0]
        )
        crystal.material = "LYSO"

        source = sim.add_source("GenericSource", "b2b")
        source.particle = "back_to_back"
        source.activity = 10 * Bq
        source.position.type = "sphere"
        source.position.radius = 0.5 * mm
        source.direction.type = "iso"
        source.direction.theta = [90, 90 * deg]
        source.direction.phi = [0, 0 * deg]

        sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"

        stats = sim.add_actor("SimulationStatisticsActor", "stats")

        hc = sim.add_actor("DigitizerHitsCollectionActor", "hits")
        hc.attached_to = crystal.name
        hc.authorize_repeated_volumes = True
        hc.output_filename = "sim.root"
        hc.attributes = [
            "EventID",
            "PreStepUniqueVolumeID",
            "TrackVolumeName",
            "PostPosition",
            "TotalEnergyDeposit",
            "GlobalTime",
        ]

        sc = sim.add_actor("DigitizerAdderActor", "singles")
        sc.attached_to = hc.attached_to
        sc.input_digi_collection = hc.name
        sc.policy = adder_policy
        sc.output_filename = hc.output_filename

        sim.run_timing_intervals = [[0, 1 * sec]]

        sim.run(start_new_process=True)

        root_file = uproot.open(hc.get_output_path())
        singles_tree = root_file["singles"]
        num_singles = int(singles_tree.num_entries)

        coincidences = coincidences_sorter(
            singles_tree,
            time_window=2 * ns,
            min_transaxial_distance=0 * mm,
            transaxial_plane="xy",
            max_axial_distance=60 * mm,
            policy="takeAllGoods",
            chunk_size=100000,
        )
        num_coincidences = len(coincidences["GlobalTime1"])
        print(f"{num_singles} single(s), {num_coincidences} coincidence(s)")

        if num_coincidences == 0:
            is_ok = False
            exceptions.append(
                GateImplementationError(
                    f"No coincidences were detected when using a DigitizerAdderActor with policy '{adder_policy}'"
                )
            )

    utility.test_ok(is_ok, exceptions=exceptions)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import os
from opengate.tests import utility
import opengate as gate
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt, TreatmentPlanSource

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test044_pbs", "test059")
    output_path = paths.output
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123654789
    sim.random_engine = "MersenneTwister"
    # sim.number_of_threads = 16
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    mrad = gate.g4_units.mrad

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # FIRST DETECTOR
    # box
    # translation and rotation like in the Gate macro
    box1 = sim.add_volume("Box", "box1")
    box1.size = [10 * cm, 10 * cm, 105 * cm]
    box1.translation = [0 * cm, 0 * cm, 52.5 * cm]
    box1.material = "Vacuum"
    box1.color = [0, 0, 1, 1]

    # phantoms
    m = Rotation.identity().as_matrix()

    phantom = sim.add_volume("Box", "phantom_a_1")
    phantom.mother = "box1"
    phantom.size = [100 * mm, 100 * mm, 50 * mm]
    phantom.translation = [0 * mm, 0 * mm, -500 * mm]
    phantom.rotation = m
    phantom.material = "G4_AIR"
    phantom.color = [1, 0, 1, 1]

    # add dose actor
    dose = sim.add_actor("DoseActor", "doseInYZ_1")
    filename = "phantom_a_1.mhd"
    dose.output_filename = filename
    dose.attached_to = "phantom_a_1"
    dose.size = [250, 250, 1]
    dose.spacing = [0.4, 0.4, 2]
    dose.hit_type = "random"

    # SECOND DETECTOR
    # box
    # translation and rotation like in the Gate macro
    box2 = sim.add_volume("Box", "box2")
    box2.size = [10 * cm, 10 * cm, 105 * cm]
    box2.translation = [30 * cm, 0 * cm, 52.5 * cm]
    box2.material = "Vacuum"
    box2.color = [0, 0, 1, 1]

    # phantoms
    m = Rotation.identity().as_matrix()

    phantom2 = sim.add_volume("Box", "phantom_a_2")
    phantom2.mother = "box2"
    phantom2.size = [100 * mm, 100 * mm, 50 * mm]
    phantom2.translation = [0 * mm, 0 * mm, -500 * mm]
    phantom2.rotation = m
    phantom2.material = "G4_AIR"
    phantom2.color = [1, 0, 1, 1]

    # add dose actor
    dose2 = sim.add_actor("DoseActor", "doseInYZ_2")
    filename = "phantom_a_2.mhd"
    dose2.output_filename = filename
    dose2.attached_to = "phantom_a_2"
    dose2.size = [250, 250, 1]
    dose2.spacing = [0.4, 0.4, 2]
    dose2.hit_type = "random"

    # TPS SOURCE
    # beamline model
    beamline = BeamlineModel()
    beamline.name = None
    beamline.radiation_types = "proton"

    # polinomial coefficients
    beamline.energy_mean_coeffs = [1, 0]
    beamline.energy_spread_coeffs = [0.4417036946562556]
    beamline.sigma_x_coeffs = [2.3335754]
    beamline.theta_x_coeffs = [2.3335754e-3]
    beamline.epsilon_x_coeffs = [0.00078728e-3]
    beamline.sigma_y_coeffs = [1.96433431]
    beamline.theta_y_coeffs = [0.00079118e-3]
    beamline.epsilon_y_coeffs = [0.00249161e-3]

    # tps
    nSim = 80000  # particles to simulate per beam

    tps = sim.add_source("TreatmentPlanPBSource", "TPSource")
    tps.n = nSim
    tps.beam_model = beamline
    tps.plan_path = ref_path / "TreatmentPlan2Spots.txt"
    tps.beam_nr = 1
    tps.gantry_rot_axis = "x"
    # tps.sorted_spot_gneration = True
    tps.particle = "proton"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.set_production_cut("world", "all", 1000 * km)
    # sim.set_user_limits("phantom_a_2","max_step_size",1,['proton'])

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # ----------------------------------------------------------------------------------------------------------------
    # tests

    # energy deposition: we expect the edep from source two
    # to be double the one of source one

    print("Compare tps Edep to single pb sources")
    print(" --------------------------------------- ")
    mhd_1 = dose.edep.get_output_path()
    mhd_2 = dose2.edep.get_output_path()
    test = True

    # check first spot
    test = (
        utility.assert_images(
            ref_path / mhd_1,
            output_path / mhd_1,
            stats,
            tolerance=70,
            ignore_value_data2=0,
        )
        and test
    )

    # check second spot
    test = (
        utility.assert_images(
            ref_path / mhd_1,
            output_path / mhd_1,
            stats,
            tolerance=70,
            ignore_value_data2=0,
        )
        and test
    )
    print(" --------------------------------------- ")
    # fig1 = utility.create_2D_Edep_colorMap(output_path / mhd_1, show=True)
    # fig2 = utility.create_2D_Edep_colorMap(output_path / mhd_2, show=True)

    print("Compare ratio of the two spots with expected ratio")

    # Total Edep
    is_ok = (
        utility.test_weights(
            2,
            output_path / mhd_1,
            output_path / mhd_2,
            thresh=0.2,
        )
        and test
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
from opengate.image import *
import opengate.contrib.spect.siemens_intevo as intevo
from opengate.contrib.spect.spect_config import SPECTConfig
import opengate as gate


def create_test_spect_config(paths):
    data_path = paths.data
    mm = gate.g4_units.mm

    sc = SPECTConfig()
    sc.simu_name = "test090"
    sc.output_folder = paths.output
    # detectors
    sc.detector_config.model = "intevo"
    sc.detector_config.collimator = "melp"
    sc.detector_config.number_of_heads = 2
    sc.detector_config.digitizer_function = intevo.add_digitizer
    sc.detector_config.size = [256, 256]
    sc.detector_config.spacing = [2.39759994 * mm, 2.39759994 * mm]
    # phantom
    sc.phantom_config.image = data_path / "iec_5mm.mhd"
    sc.phantom_config.labels = data_path / "iec_5mm_labels.json"
    sc.phantom_config.material_db = data_path / "iec_5mm.db"
    # source
    sc.source_config.image = data_path / "iec_5mm_activity.mhd"
    sc.source_config.radionuclide = "177lu"
    sc.source_config.total_activity = 1e3 * gate.g4_units.Bq
    # acquisition
    sc.acquisition_config.radius = 350 * gate.g4_units.mm
    sc.acquisition_config.duration = 30 * gate.g4_units.s
    sc.acquisition_config.number_of_angles = 3
    return sc


def check_stats_file(n, sc, stats, is_ok):
    print(stats.get_output_path())
    stats = utility.read_stat_file(stats.get_output_path())
    print(stats)
    b = stats.counts.runs == sc.acquisition_config.number_of_angles
    is_ok = is_ok and b
    utility.print_test(
        b,
        f"Number of runs {sc.acquisition_config.number_of_angles} vs {stats.counts.runs}",
    )
    b = n * 0.9 < stats.counts.events < n * 1.1
    is_ok = is_ok and b
    utility.print_test(b, f"Number of events {n} vs {stats.counts.events}")
    return is_ok


def check_projection_files(sim, paths, stats, is_ok, tol=60, squared_flag=False):
    stats = utility.read_stat_file(stats.get_output_path())
    # check images
    channel = sim.actor_manager.find_actors("energy_window")[0]
    projs = sim.actor_manager.find_actors("projection")
    for d in projs:
        # counts
        ref_file = paths.output_ref / "projection_0_counts.mhd"
        img_file = d.get_output_path("counts")
        b = utility.assert_images(
            ref_file,
            img_file,
            tolerance=1000,
            ignore_value_data1=0,
            ignore_value_data2=0,
            sum_tolerance=tol,
        )
        is_ok = is_ok and b

        info = read_image_info(img_file)
        print(info.size, info.spacing)
        b = info.size[2] == len(channel.channels) * stats.counts.runs
        utility.print_test(
            b, f"Number of channels {len(channel.channels)}x{stats.counts.runs}"
        )
        is_ok = is_ok and b

        # squared
        if squared_flag:
            img_file = d.get_output_path("squared_counts")
            info = read_image_info(img_file)
            print(info.size, info.spacing)
            b = info.size[2] == len(channel.channels) * stats.counts.runs
            utility.print_test(
                b, f"Number of channels {len(channel.channels)}x{stats.counts.runs}"
            )
            is_ok = is_ok and b

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test028_ge_nm670_spect_2_helpers as test028
from opengate.tests import utility


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test028_hits")

    # create the simulation
    sim = gate.Simulation()

    # main description
    test028.create_spect_simu(sim, paths)

    # mono thread
    sim.number_of_threads = 1

    sim.run()

    # check
    is_ok = test028.test_spect_root(sim, paths)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4
import uproot
import numpy as np
from scipy.spatial.transform import Rotation
from opengate.tests import utility


def check_process_user_hook(simulation_engine):
    # Check whether the particle 'gamma' actually has
    # the requested processes attached to it
    p_name = "e-"
    g4_particle_table = g4.G4ParticleTable.GetParticleTable()
    particle = g4_particle_table.FindParticle(particle_name=p_name)
    # FindParticle returns nullptr if particle name was not found
    if particle is None:
        raise Exception(f"Something went wrong. Could not find particle {p_name}.")
    pm = particle.GetProcessManager()
    process = "eBrem"
    process_list = pm.GetProcessList()
    isInside = False
    for i in range(process_list.size()):
        processName = str(process_list[i].GetProcessName())
        if "biasWrapper(" + process + ")" == processName:
            isInside = True

    # GetProcess returns nullptr if the requested process was not found
    if isInside == False:
        raise Exception(
            f"Could not find the process '{process}' for particle {p_name}."
        )
    else:
        print(f"Hooray, I found the process '{process}' for the particle {p_name}!")


def validation_test(arr, nb_split, tol=0.04):
    arr = arr[arr["ParticleName"] == "gamma"]
    EventID = arr["EventID"]
    Weights = arr["Weight"][EventID == EventID[0]]
    sum_Weights = np.round(np.sum(Weights), 4)
    if 1 - tol < sum_Weights < 1 + tol and nb_split * (1 - tol) < len(
        Weights
    ) < nb_split * (1 + tol):
        return True
    else:
        print("Test failed:")
        print(f"{sum_Weights=} vs 1, tol={tol}")
        print(f"{Weights=}  {len(Weights)=}")
        print(f"{nb_split=}")
        return False


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "test070test_operator_brem_splitting", output_folder="test070"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    # sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.EVENT
    sim.number_of_threads = 1
    sim.random_seed = 123456789
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    km = gate.g4_units.km
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    gcm3 = gate.g4_units.g / gate.g4_units.cm3

    #  adapt world size
    sim.world.size = [1 * m, 1 * m, 2 * m]
    sim.world.material = "G4_Galactic"

    ####### GEOMETRY TO IRRADIATE #############
    sim.volume_manager.material_database.add_material_weights(
        "Tungsten",
        ["W"],
        [1],
        19.3 * gcm3,
    )

    W_tubs = sim.add_volume("Tubs", "W_box")
    W_tubs.material = "Tungsten"

    W_tubs.rmin = 0
    W_tubs.rmax = 0.1 * um
    W_tubs.dz = 0.5 * m
    W_tubs.color = [0.8, 0.2, 0.1, 1]
    angle_x = np.random.randint(0, 360)
    angle_y = np.random.randint(0, 360)
    angle_z = np.random.randint(0, 360)
    # angle_x = 45
    # angle_y = 70
    # angle_z = 80

    rotation = Rotation.from_euler(
        "xyz", [angle_y, angle_y, angle_z], degrees=True
    ).as_matrix()
    W_tubs.rotation = rotation

    ######## BremSplitting ACTOR #########
    nb_split = 100
    brem_splitting_actor = sim.add_actor(
        "BremsstrahlungSplittingActor", "eBremSplittingW"
    )
    brem_splitting_actor.attached_to = W_tubs.name
    brem_splitting_actor.splitting_factor = nb_split
    brem_splitting_actor.particles = "e-", "e+"

    ##### PHASE SPACE plan ######"
    plan_tubs = sim.add_volume("Tubs", "phsp_tubs")
    plan_tubs.material = "G4_Galactic"
    plan_tubs.rmin = 0.11 * um
    plan_tubs.rmax = 0.11 * um + 1 * nm
    plan_tubs.dz = 0.5 * m
    plan_tubs.color = [0.2, 1, 0.8, 1]
    plan_tubs.rotation = rotation

    ####### Electron source ###########
    source = sim.add_source("GenericSource", "source1")
    source.particle = "e-"
    source.n = 10000
    source.position.type = "sphere"
    source.position.radius = 1 * nm
    source.direction.type = "momentum"
    source.direction.momentum = np.dot(rotation, np.array([0, 0, 1]))
    source.energy.type = "mono"
    source.energy.mono = 6 * MeV

    ####### PHASE SPACE ACTOR ##############

    phsp_actor = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phsp_actor.attached_to = plan_tubs
    phsp_actor.attributes = [
        "EventID",
        "Weight",
        "ParticleName",
    ]

    phsp_actor.output_filename = "test070_output_data.root"

    ##### MODIFIED PHYSICS LIST ###############

    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"

    sim.physics_manager.global_production_cuts.gamma = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 1 * um
    sim.physics_manager.global_production_cuts.positron = 1 * km

    sim.user_hook_after_init = check_process_user_hook

    sim.run()

    #
    # # print results
    print(stats)
    #
    f_phsp = uproot.open(phsp_actor.get_output_path())
    arr = f_phsp["PhaseSpace"].arrays()

    is_ok = validation_test(arr, nb_split)
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test036_adder_depth_helpers as t036
from opengate.tests import utility
from opengate import g4_units

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test036_adder_depth", "test036"
    )

    # create and run the simulation
    mm = g4_units.mm
    sim = t036.create_simulation("param", paths, "_notattached")

    # sim.visu = True
    sim.visu_type = "qt"

    # remove one head
    head = sim.volume_manager.get_volume("SPECThead")
    head.translation = head.translation[0]
    head.rotation = head.rotation[0]

    # test another case that should fail
    proj2 = sim.add_actor("DigitizerProjectionActor", "proj2")
    proj2.attached_to = "crystal_pixel"
    proj2.output_filename = "proj2-onlyexistsiffail.mha"
    proj2.size = [128, 128]
    proj2.spacing = [5 * mm, 5 * mm]

    is_ok = True
    try:
        sim.run(start_new_process=True)
        is_ok = False
    except:
        pass
    utility.print_test(is_ok, f"The run should NOT be executed: ")
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import uproot
import numpy as np

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test023")

    # create the simulation
    sim = gate.Simulation()
    sim_name = "test023_filters_attribute_bool"

    # main options
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.random_seed = 321456987
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    mm = gate.g4_units.mm
    sec = gate.g4_units.second
    keV = gate.g4_units.keV

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]
    sim.world.material = "G4_WATER"

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.type = "gauss"
    source.energy.mono = 1 * MeV
    source.energy.sigma_gauss = 1 * MeV
    source.particle = "gamma"
    source.position.type = "sphere"
    source.position.radius = 1 * cm
    source.direction.type = "iso"
    source.activity = 100 * Bq

    # plane
    plane1a = sim.add_volume("Box", "plane1a")
    plane1a.size = [1 * m, 1 * m, 1 * nm]
    plane1a.translation = [0, 0, 1 * cm]
    plane1a.material = "G4_WATER"
    plane1a.color = [1, 0, 0, 1]

    # filter according to time
    filter1 = sim.add_filter("ThresholdAttributeFilter", "time_filter")
    filter1.attribute = "GlobalTime"
    filter1.value_min = 20 * sec
    filter1.value_max = 70 * sec
    filter1.policy = "accept"

    # filter according to energy
    filter2 = sim.add_filter("ThresholdAttributeFilter", "ene_filter")
    filter2.attribute = "KineticEnergy"
    filter2.value_min = 300 * keV
    filter2.value_max = 1200 * keV
    filter2.policy = "accept"

    # filter according to particle
    filter3 = sim.add_filter("ParticleFilter", "p_filter")
    filter3.particle = "gamma"

    # phsp
    phsp_and = sim.add_actor("PhaseSpaceActor", "phsp_and")
    phsp_and.attached_to = plane1a.name
    phsp_and.attributes = ["GlobalTime", "KineticEnergy", "ParticleName"]
    phsp_and.output_filename = f"{sim_name}_and.root"
    phsp_and.filters = [filter1, filter2, filter3]
    phsp_and.filters_boolean_operator = "and"  # default is and

    # phsp
    phsp_or = sim.add_actor("PhaseSpaceActor", "phsp_or")
    phsp_or.attached_to = plane1a.name
    phsp_or.attributes = ["GlobalTime", "KineticEnergy"]
    phsp_or.output_filename = f"{sim_name}_or.root"
    phsp_or.filters.append(filter1)
    phsp_or.filters.append(filter2)
    phsp_or.filters_boolean_operator = "or"  # default is and

    # stats
    stat = sim.add_actor("SimulationStatisticsActor", "stats")
    stat.track_types_flag = True

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"

    # start simulation
    duration = 100 * sec
    sim.run_timing_intervals = [[0, duration]]
    sim.run()

    # print results at the end
    print(stat)
    # reference :
    # stat.write(paths.output_ref / f"{sim_name}.txt")

    # check 'or'
    print()
    print()
    tree = uproot.open(phsp_and.get_output_path())["phsp_and"]
    print("nb entries", tree.num_entries)
    ene = tree.arrays(["GlobalTime", "KineticEnergy"])["KineticEnergy"]
    emin = np.min(ene)
    emax = np.max(ene)
    is_ok = emin >= filter2.value_min and emax <= filter2.value_max
    utility.print_test(is_ok, f"Ene = {len(ene)} min={emin/keV} max={emax/keV}")
    ti = tree.arrays(["GlobalTime", "KineticEnergy"])["GlobalTime"]
    tmin = np.min(ti)
    tmax = np.max(ti)
    is_ok = tmin >= filter1.value_min and tmax <= filter1.value_max and is_ok
    utility.print_test(is_ok, f"Time = {len(ti)} min={tmin/sec} max={tmax/sec}")

    # check 'or'
    print()
    print()
    tree = uproot.open(phsp_or.get_output_path())["phsp_or"]
    print("nb entries", tree.num_entries)
    ene = tree.arrays(
        ["GlobalTime", "KineticEnergy"],
        f"(GlobalTime <= {filter1.value_min}) | "
        f"(GlobalTime >= {filter1.value_max})",
    )["KineticEnergy"]
    emin = np.min(ene)
    emax = np.max(ene)
    is_ok = emin >= filter2.value_min and emax <= filter2.value_max
    utility.print_test(is_ok, f"Ene = {len(ene)} min={emin/keV} max={emax/keV}")
    ti = tree.arrays(
        ["GlobalTime", "KineticEnergy"],
        f"(KineticEnergy <= {filter2.value_min}) | "
        f"(KineticEnergy >= {filter2.value_max})",
    )["GlobalTime"]
    tmin = np.min(ti)
    tmax = np.max(ti)
    is_ok = tmin >= filter1.value_min and tmax <= filter1.value_max and is_ok
    utility.print_test(is_ok, f"Time = {len(ti)} min={tmin/sec} max={tmax/sec}")

    # tests
    print()
    print()
    stats_ref = utility.read_stat_file(paths.output_ref / f"{sim_name}.txt")
    is_ok = utility.assert_stats(stat, stats_ref, 0.01)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect.ge_discovery_nm670 as nm670
from opengate.userhooks import check_production_cuts
from opengate.tests import utility


def create_spect_simu(sim, paths, number_of_threads=1):
    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "qt"
    sim.number_of_threads = number_of_threads
    sim.random_seed = 12345678
    # sim.random_seed = "auto"
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    # world size
    sim.world.size = [1 * m, 1 * m, 1 * m]
    sim.world.material = "G4_AIR"

    # spect head (debug mode = very small collimator)
    spect, colli, crystal = nm670.add_spect_head(
        sim, "spect", collimator_type=False, debug=False
    )
    nm670.rotate_gantry(spect, 20 * cm, 0)

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [15 * cm, 15 * cm, 15 * cm]
    waterbox.material = "G4_WATER"
    blue = [0, 1, 1, 1]
    waterbox.color = blue

    # physic list
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = False

    sim.physics_manager.global_production_cuts.gamma = 10 * mm
    sim.physics_manager.global_production_cuts.electron = 10 * mm
    sim.physics_manager.global_production_cuts.positron = 10 * mm
    sim.physics_manager.global_production_cuts.proton = 10 * mm

    sim.physics_manager.set_production_cut(
        volume_name="spect",
        particle_name="gamma",
        value=0.1 * mm,
    )
    sim.physics_manager.set_production_cut(
        volume_name="spect",
        particle_name="electron",
        value=0.01 * mm,
    )
    sim.physics_manager.set_production_cut(
        volume_name="spect",
        particle_name="positron",
        value=0.1 * mm,
    )

    # default source for tests
    activity = 30 * kBq
    beam1 = sim.add_source("GenericSource", "beam1")
    beam1.attached_to = waterbox.name
    beam1.particle = "gamma"
    beam1.energy.mono = 140.5 * keV
    beam1.position.type = "sphere"
    beam1.position.radius = 3 * cm
    beam1.position.translation = [0, 0, 0 * cm]
    beam1.direction.type = "momentum"
    beam1.direction.momentum = [0, 1, 0]
    # beam1.direction.type = 'iso'
    beam1.activity = activity / sim.number_of_threads

    beam2 = sim.add_source("GenericSource", "beam2")
    beam2.attached_to = waterbox.name
    beam2.particle = "gamma"
    beam2.energy.mono = 140.5 * keV
    beam2.position.type = "sphere"
    beam2.position.radius = 3 * cm
    beam2.position.translation = [18 * cm, 0, 0]
    beam2.direction.type = "momentum"
    beam2.direction.momentum = [0, 1, 0]
    # beam2.direction.type = 'iso'
    beam2.activity = activity / sim.number_of_threads

    beam3 = sim.add_source("GenericSource", "beam3")
    beam3.attached_to = waterbox.name
    beam3.particle = "gamma"
    beam3.energy.mono = 140.5 * keV
    beam3.position.type = "sphere"
    beam3.position.radius = 1 * cm
    beam3.position.translation = [0, 10 * cm, 0]
    beam3.direction.type = "momentum"
    beam3.direction.momentum = [0, 1, 0]
    # beam3.direction.type = 'iso'
    beam3.activity = activity / sim.number_of_threads

    # add stat actor
    stats_actor = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats_actor.track_types_flag = True
    stats_actor.output_filename = "stats.txt"

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    # get crystal volume by looking for the word crystal in the name
    for k, v in sim.volume_manager.volumes.items():
        if "crystal" in k:
            crystal = v
    hc.attached_to = crystal.name
    print("Crystal :", crystal.name)
    hc.output_filename = "test028.root"
    print(hc.output_filename)
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "TrackVolumeCopyNo",
        "PostStepUniqueVolumeID",
        "PreStepUniqueVolumeID",
        "GlobalTime",
        "KineticEnergy",
        "ProcessDefinedStep",
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.attached_to = crystal.name
    sc.input_digi_collection = "Hits"
    sc.policy = "EnergyWinnerPosition"
    # sc.policy = 'EnergyWeightedCentroidPosition'
    sc.skip_attributes = ["KineticEnergy", "ProcessDefinedStep"]
    sc.output_filename = hc.output_filename

    # EnergyWindows
    cc = sim.add_actor("DigitizerEnergyWindowsActor", "EnergyWindows")
    cc.attached_to = crystal.name
    cc.input_digi_collection = "Singles"
    cc.channels = [
        {"name": "scatter", "min": 114 * keV, "max": 126 * keV},
        {"name": "peak140", "min": 126 * keV, "max": 154.55 * keV},
        {
            "name": "spectrum",
            "min": 0 * keV,
            "max": 5000 * keV,
        },  # should be strictly equal to 'Singles'
    ]
    cc.output_filename = hc.output_filename

    """
        The order of the actors is important !
        1. Hits
        2. Singles
        3. EnergyWindows
    """

    # sec = gate.g4_units.second
    # sim.run_timing_intervals = [[0, 0.5 * sec], [0.5 * sec, 1 * sec]]

    # user hook function
    sim.user_hook_after_init = check_production_cuts

    return spect


def test_add_proj(sim):
    mm = gate.g4_units.mm
    crystal = None
    for k, v in sim.volume_manager.volumes.items():
        if "crystal" in k:
            crystal = v
    # 2D binning projection
    proj = sim.add_actor("DigitizerProjectionActor", "Projection")
    proj.attached_to = crystal.name
    # we set two times the spectrum channel to compare with Gate output
    proj.input_digi_collections = ["spectrum", "scatter", "peak140", "spectrum"]
    proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.size = [128, 128]
    proj.output_filename = f"proj028.mhd"
    # by default, the origin of the images are centered
    # set to False here to keep compatible with previous version
    # proj.origin_as_image_center = False
    return proj


def test_spect_root(sim, paths):
    # stat
    gate.exception.warning("Compare stats")
    stats = sim.get_actor("Stats")
    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "stats.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.07)

    # Compare root files
    print()
    gate.exception.warning("Compare hits")
    ref_file = paths.output_ref / "test028.root"
    hc_file = sim.get_actor("Hits").get_output_path()
    print(ref_file, hc_file)
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 1.7, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 1.3, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.9, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.001,
            "scaling": 1,
        },
        {"k1": "GlobalTime", "k2": "GlobalTime", "tol": 1.5e7, "scaling": 1},
    ]
    is_ok = (
        utility.compare_root2(
            ref_file,
            hc_file,
            "Hits",
            "Hits",
            checked_keys,
            paths.output / "test028_hits.png",
            n_tol=4,
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.exception.warning("Compare singles")
    hc_file = sim.get_actor("Singles").get_output_path()
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 1.8, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 1.3, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.6, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.001,
            "scaling": 1,
        },
    ]
    is_ok = (
        utility.compare_root2(
            ref_file,
            hc_file,
            "Singles",
            "Singles",
            checked_keys,
            paths.output / "test028_singles.png",
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.exception.warning("Compare singles and spectrum (must be strictly equal)")
    ref_file = sim.get_actor("Singles").get_output_path()
    hc_file = sim.get_actor("EnergyWindows").get_output_path()
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 0.001, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 0.001, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.001, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.001,
            "scaling": 1,
        },
    ]
    is_ok = (
        utility.compare_root2(
            ref_file,
            hc_file,
            "Singles",
            "spectrum",
            checked_keys,
            paths.output / "test028_spectrum.png",
            n_tol=0.01,
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.exception.warning("Compare scatter")
    ref_file = paths.output_ref / "test028.root"
    hc_file = sim.get_actor("EnergyWindows").get_output_path()
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 13, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 5, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 8, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.2,
            "scaling": 1,
        },
    ]
    is_ok = (
        utility.compare_root2(
            ref_file,
            hc_file,
            "scatter",
            "scatter",
            checked_keys,
            paths.output / "test028_scatter.png",
            n_tol=15,
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.exception.warning("Compare peak")
    ref_file = paths.output_ref / "test028.root"
    hc_file = sim.get_actor("EnergyWindows").get_output_path()
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 1.1, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 0.4, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.4, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.1,
            "scaling": 1,
        },
    ]
    is_ok = (
        utility.compare_root2(
            ref_file,
            hc_file,
            "peak140",
            "peak140",
            checked_keys,
            paths.output / "test028_peak.png",
            n_tol=2.1,
        )
        and is_ok
    )

    return is_ok


def test_spect_proj(sim, paths, proj, output_ref_folder=None, output_ref_filename=None):
    print()
    stats = sim.get_actor("Stats")
    stats.user_output.stats.merged_data.runs = 1  # force to 1 to compare with MT
    print(stats)
    if output_ref_folder is None:
        output_ref_folder = paths.output_ref
    stats_ref = utility.read_stat_file(output_ref_folder / "stats.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.025)

    # compare images with Gate
    print()
    print("Compare images with ref")
    fn = proj.get_output_path("counts")
    if output_ref_filename is None:
        output_ref_filename = paths.output_ref / fn.name
    print(fn)
    print(output_ref_filename)
    is_ok = (
        utility.assert_images(
            output_ref_folder / output_ref_filename,
            fn,
            stats,
            tolerance=16,
            ignore_value_data2=0,
            axis="x",
            fig_name=paths.output / f"{fn.stem}.png",
            sum_tolerance=6,
        )
        and is_ok
    )
    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test019_linac_phsp_helpers as test019
import opengate as gate

if __name__ == "__main__":
    sim = test019.init_test019(1, "linacphsp_visu")

    sim.visu = True
    sim.visu_type = "vrml"
    sim.visu_filename = "geant4VisuFile.wrl"

    source = sim.get_source_user_info("Default")
    source.activity = 1 * gate.g4_units.Bq

    print(sim.volume_manager.dump_volume_tree())

    test019.run_test019(sim)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import click
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate import logger
from opengate.managers import Simulation
from opengate.utility import g4_units
from opengate.image import *
from opengate.voxelize import write_voxelized_geometry, voxelized_source

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("--spacing", "-s", default=4.0, help="Spacing in mm")
@click.option("--output", "-o", required=True, help="output filename (.mhd)")
@click.option(
    "--output_source", default=None, help="output filename for vox source (.mhd)"
)
@click.option(
    "--activities",
    "-a",
    default=None,
    nargs=6,
    help="List of 6 activities for the 6 spheres: 10, 13, 17, 22, 28, 37",
)
@click.option("--bg", default=0.0, help="Activity in the background")
@click.option("--cyl", default=0.0, help="Activity in the central cylinder")
@click.option(
    "--no_shell",
    is_flag=True,
    default=False,
    help="If set, do not consider the shell of the sphere (for high resolution)",
)
def go(output, spacing, output_source, activities, no_shell, bg, cyl):
    # create the simulation
    sim = Simulation()
    sim.verbose_level = logger.INFO

    # world
    m = g4_units.m
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a iec phantom
    iec = gate_iec.add_iec_phantom(
        sim, sphere_starting_angle=False, toggle_sphere_order=False
    )

    # voxelized the iec volume
    print("Starting voxelization ...")
    spacing = (spacing, spacing, spacing)
    volume_labels, image = sim.voxelize_geometry(extent=iec, spacing=spacing, margin=1)

    info = get_info_from_image(image)
    print(f"Image size={info.size}")
    print(f"Image spacing={info.spacing}")
    print(f"Image origin={info.origin}")

    # write files
    filenames = write_voxelized_geometry(sim, volume_labels, image, output)
    for f in filenames.values():
        print(f"Output: {f}")

    # voxelized source activities
    if activities is None:
        activities = [0.0] * 6
    a = {
        "iec_sphere_10mm": activities[0],
        "iec_sphere_13mm": activities[1],
        "iec_sphere_17mm": activities[2],
        "iec_sphere_22mm": activities[3],
        "iec_sphere_28mm": activities[4],
        "iec_sphere_37mm": activities[5],
        "iec_interior": bg,
        "iec_center_cylinder_hole": cyl,
    }
    if not no_shell:
        a["iec_sphere_shell_10mm"] = activities[0]
        a["iec_sphere_shell_13mm"] = activities[1]
        a["iec_sphere_shell_17mm"] = activities[2]
        a["iec_sphere_shell_22mm"] = activities[3]
        a["iec_sphere_shell_28mm"] = activities[4]
        a["iec_sphere_shell_37mm"] = activities[5]

    if output_source is not None:
        itk_source = voxelized_source(image, volume_labels, a)
        print(f"Write image source {output_source}")
        itk.imwrite(itk_source, output_source)


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
import itk
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from matplotlib.animation import FuncAnimation, PillowWriter
from PIL import Image
from sklearn.mixture import GaussianMixture
from scipy.stats import multivariate_normal


def main():
    read_pb_positions_and_weights_from_file = True
    default_dir = Path("__file__").parent  # Path.cwd()
    if read_pb_positions_and_weights_from_file:
        file_path_spot_weights = default_dir / "gate_logo_gmm_results.txt"
    else:
        image_path = default_dir / "GATE_logo.png"  # Replace with your image path
        output_file = default_dir / "gate_logo_gmm_results_newlyoptimized.txt"
        n_gaussians_to_represent_img = 150
        convert_img_to_gaussian_mixture(
            image_path, output_file, n_components=n_gaussians_to_represent_img
        )
        file_path_spot_weights = output_file

    expt_fpath = default_dir / Path("Gate_logo_simulated.png")

    run_simulation_create_gif(file_path_spot_weights, expt_fpath)


def run_simulation_create_gif(
    file_path_spot_weights,
    expt_fpath,
    number_of_gif_frames=20,
    activity=5e7,
    sigma_smearing_factor=1.05,
):
    multiple_runs = True
    weight_source_activity = True  # if False use about 1e6 [Bq]

    # create the simulation
    sim = gate.Simulation()
    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 1234567891
    sim.number_of_threads = 1
    sim.output_dir = expt_fpath.parent / "temp"
    if not multiple_runs:
        numPartSimTest = 4e5 / sim.number_of_threads

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    s = gate.g4_units.s

    #  change world size
    world = sim.world
    world.size = [60 * cm, 50 * cm, 50 * cm]
    # world.material = "Vacuum"

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [10 * cm, 40 * cm, 20 * cm]
    phantom.translation = [-5 * cm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMV"

    M_spots = np.genfromtxt(file_path_spot_weights, delimiter="\t")

    w_sum = np.sum(M_spots[:, 2])
    ## normalize weights in case is not normalized
    M_spots[:, 2] = M_spots[:, 2] / w_sum

    m = np.amax(M_spots, axis=0)

    sorted_indices = np.lexsort((M_spots[:, 0], M_spots[:, 1]))
    M_spots = M_spots[sorted_indices]
    # print(f'{w_sum = }')
    # print(M_sum)
    # print(f'{m=}')
    # print(f'{M_spots.shape = }')
    n_sources = M_spots.shape[0]
    id_interv = n_sources // number_of_gif_frames
    # print(f'{id_interv = }')

    meterset_weight_sum = 0.0
    run_time_intervals = []
    for i, row in enumerate(M_spots):
        x_pos = float(row[0]) - m[0] / 2
        y_pos = float(row[1]) - m[1] / 2
        w = float(row[2])

        sigma = float(row[3]) * sigma_smearing_factor
        source = sim.add_source("GenericSource", f"mysource_{i}")
        source.energy.mono = 1 * MeV
        source.particle = "proton"
        source.position.type = "disc"
        # print(dir(source.position))
        source.position.rotation = Rotation.from_euler(
            "y", 90, degrees=True
        ).as_matrix()
        # sigma = 8
        source.position.sigma_x = sigma * mm
        source.position.sigma_y = sigma * mm
        source.direction.type = "momentum"
        source.direction.momentum = [-1, 0, 0]
        source.position.translation = [0 * mm, y_pos * mm, x_pos * mm]

        # print(dir(source.energy))
        if multiple_runs:
            """There is two ways of setting the weight of the sources correctly (both work)"""

            if weight_source_activity:
                source.activity = w * activity * Bq
                source.start_time = (i) / n_sources * s
                source.end_time = (i + 1) / n_sources * s
            else:
                source.activity = activity * Bq
                source.start_time = meterset_weight_sum * s
                meterset_weight_sum += w
                source.end_time = meterset_weight_sum * s

            if (i % id_interv) == 0 or (i == (n_sources - 1)):
                if i == 0:
                    start_time = 0 * s
                else:
                    start_time = run_time_intervals[-1][1]
                run_time_intervals.append([start_time, source.end_time])

        else:
            source.n = np.ceil(numPartSimTest * w)

    size = [1, 400, 200]
    spacing = [100 * mm, 1.0 * mm, 1.0 * mm]

    doseActorName = "gate_logo"
    doseActor = sim.add_actor("DoseActor", doseActorName)
    doseActor.output_filename = "test0-" + doseActorName + ".mhd"
    doseActor.attached_to = phantom
    doseActor.size = size
    doseActor.spacing = spacing
    doseActor.hit_type = "random"
    doseActor.dose.active = False
    if multiple_runs:
        doseActor.keep_data_per_run = True
        sim.run_timing_intervals = run_time_intervals
        print(run_time_intervals)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    sim.run()

    fname1 = str(doseActor.edep.get_output_path())
    total_dose_all_runs = itk.imread(fname1)

    data_array_all_runs = np.squeeze(itk.GetArrayViewFromImage(total_dose_all_runs))
    max_value_data_arr = np.amax(data_array_all_runs[:])
    plt.imshow(data_array_all_runs)
    plt.imsave(
        expt_fpath.with_stem(f"{expt_fpath.stem}_static").with_suffix(".png"),
        data_array_all_runs,
    )
    plt.show()

    if multiple_runs:
        fpath_full_img = Path(doseActor.edep.get_output_path())
        frames = []
        data_array_sum = np.zeros_like(
            data_array_all_runs, dtype=data_array_all_runs.dtype
        )
        for i in range(len(run_time_intervals)):
            f_path_i = fpath_full_img.with_stem(f"{fpath_full_img.stem}-run{i}")
            img_i = itk.imread(str(f_path_i))
            data_array_sum += np.squeeze(itk.GetArrayViewFromImage(img_i))
            # we normalize the data to their cumulative max value, so data range from 0 to 1
            frames.append(data_array_sum.copy() / max_value_data_arr)
        # Create the figure and axis
        fig, ax = plt.subplots()
        im = ax.imshow(frames[0], cmap="viridis", aspect="auto", vmin=-0, vmax=1)
        ax.set_title("Frame 0")

        # Update function for each frame
        def update(frame):
            im.set_data(frames[frame])
            ax.set_title(f"Frame {frame}")
            return im, ax

        # Create the animation
        ani = FuncAnimation(
            fig, update, frames=len(frames), interval=100, repeat_delay=1000
        )

        # Save the animation as a GIF
        writer = PillowWriter(fps=5)  # 5 frames per second
        ani.save(
            expt_fpath.with_stem(f"{expt_fpath.stem}_animated").with_suffix(".gif"),
            writer=writer,
        )


def load_and_discretize_image(image_path, sigma_blurr=0, levels=5):
    img = Image.open(image_path).convert("L")
    image_array = np.array(img)
    # Apply Gaussian filter (blur)
    # image_array = gaussian_filter(image_array, sigma=sigma_blurr)
    # Normalize the image and scale it to the desired levels
    max_val = 255  # Maximum grayscale value
    step = max_val // levels
    # Discretize and preserve the original contrast
    discretized = (max_val - image_array) // step
    return discretized


# Perform Gaussian Mixture analysis
def analyze_with_gmm_discretized(image_array, n_components=10):
    # Get coordinates and intensity values
    coords = np.column_stack(np.where(image_array > 0))
    intensities = image_array[coords[:, 0], coords[:, 1]]

    # Repeat coordinates based on discretized intensity values
    repeated_coords = np.repeat(coords, intensities // (1), axis=0)

    # Fit the GMM without sample_weight
    gmm = GaussianMixture(
        n_components=n_components, random_state=42, covariance_type="spherical"
    )
    # gmm = BayesianGaussianMixture(n_components=n_components,tol = 1e-4, covariance_type = 'spherical', random_state=42)
    gmm.fit(repeated_coords)
    return gmm


# Custom GMM fitting with fixed sigma
def fixed_sigma_gmm(X, n_components, sigma):
    n_samples, n_features = X.shape
    gmm = GaussianMixture(
        n_components=n_components, covariance_type="spherical", random_state=42
    )
    gmm.fit(X)

    # Override the covariances with the fixed sigma squared
    gmm.covariances_ = np.full(n_components, sigma**2)  # Variance = sigma^2
    gmm.precisions_cholesky_ = np.sqrt(
        1 / gmm.covariances_
    )  # Precomputed for faster predictions

    # Predict cluster assignments
    labels = gmm.predict(X)
    return gmm, labels


# Calculate the cumulative Gaussian map
def calculate_cumulative_gaussian_map(image_shape, gmm):
    x = np.arange(image_shape[1])
    y = np.arange(image_shape[0])
    xx, yy = np.meshgrid(x, y)
    grid_coords = np.column_stack([yy.ravel(), xx.ravel()])
    cumulative_map = np.zeros(image_shape)

    for mean, covar, weight in zip(gmm.means_, gmm.covariances_, gmm.weights_):
        gaussian = multivariate_normal(mean=mean, cov=covar)
        cumulative_map += weight * gaussian.pdf(grid_coords).reshape(image_shape)

    return cumulative_map


# Plot results
def plot_results(image_array, gmm, cumulative_map):
    fig, axes = plt.subplots(1, 2, figsize=(15, 7))

    # Plot discretized grayscale image
    axes[0].imshow(image_array, cmap="gray", origin="upper")
    centers = gmm.means_
    weights = gmm.weights_
    axes[0].scatter(
        centers[:, 1],
        centers[:, 0],
        c="red",
        s=weights * 1000,
        alpha=0.7,
        label="GMM Centers",
    )
    axes[0].set_title("GMM Analysis on Discretized Grayscale Image")
    axes[0].legend()

    # Plot cumulative Gaussian map
    axes[1].imshow(cumulative_map, cmap="hot", origin="upper")
    axes[1].set_title("Cumulative Gaussian Map")
    plt.show()


# Save results to a text file
def save_results_to_file(gmm, output_file):
    centers = gmm.means_
    weights = gmm.weights_
    sigmas = np.sqrt(gmm.covariances_)
    with open(output_file, "w") as f:
        for i, (x, y) in enumerate(centers):
            f.write(f"{x:.2f}\t{y:.2f}\t{weights[i]:.4e}\t{sigmas[i]:.4e}\n")


# Main workflow
def convert_img_to_gaussian_mixture(image_path, output_file, n_components=10, levels=5):
    image_array = load_and_discretize_image(image_path, levels)
    gmm = analyze_with_gmm_discretized(image_array, n_components)
    cumulative_map = calculate_cumulative_gaussian_map(image_array.shape, gmm)
    plot_results(image_array, gmm, cumulative_map)
    save_results_to_file(gmm, output_file)
    print(f"Results saved to {output_file}")


if __name__ == "__main__":
    main()



from xml.etree import ElementTree as ET

from box import Box

from enum import Enum
import xml.etree.ElementTree as ET

import opengate_core as g4

from .exception import warning, fatal
from .definitions import FLOAT_MAX
from .decorators import requires_fatal
from .base import GateObject, process_cls

from .utility import g4_units, get_material_name_variants

# names for particle cuts
cut_particle_names = {
    "gamma": "gamma",
    "electron": "e-",
    "positron": "e+",
    "proton": "proton",
}


# translation from particle names used in Gate
# to particles names used in Geant4
def translate_particle_name_gate_to_geant4(name):
    """Convenience function to translate from names
    used in Gate to those in G4, if necessary.
    Concerns e.g. 'electron' -> 'e-'
    """
    try:
        return cut_particle_names[name]
    except KeyError:
        return name


class UserLimitsPhysics(g4.G4VPhysicsConstructor):
    """
    Class to be registered to physics list.

    It is essentially a refined version of StepLimiterPhysics which considers the user's
    particles choice of particles to which the step limiter should be added.

    """

    def __init__(self):
        """Objects of this class are created via the PhysicsEngine class.
        The user should not create objects manually.

        """
        g4.G4VPhysicsConstructor.__init__(self, "UserLimitsPhysics")
        self.physics_engine = None

        self.g4_step_limiter_storage = {}
        self.g4_special_user_cuts_storage = {}

    def close(self):
        self.g4_step_limiter_storage = None
        self.g4_special_user_cuts_storage = None
        self.physics_engine = None

    @requires_fatal("physics_engine")
    def ConstructParticle(self):
        """Needs to be defined because C++ base class declares this as purely virtual member."""
        pass

    @requires_fatal("physics_engine")
    def ConstructProcess(self):
        """Overrides method from G4VPhysicsConstructor
        that is called when the physics list is constructed.

        """
        ui = self.physics_engine.user_info_physics_manager

        particle_keys_to_consider = []
        # 'all' overrides individual settings
        if ui.user_limits_particles["all"] is True:
            particle_keys_to_consider = list(ui.user_limits_particles.keys())
        else:
            keys_to_exclude = ("all", "all_charged")
            particle_keys_to_consider = [
                p
                for p, v in ui.user_limits_particles.items()
                if v is True and p not in keys_to_exclude
            ]

        if len(particle_keys_to_consider) == 0:
            self.physics_engine.simulation_engine.simulation.warn_user(
                "user_limits_particles is False for all particles. No tracking cuts will be applied. Use sim.physics_manager.set_user_limits_particles()."
            )

        # translate to Geant4 particle names
        particles_to_consider = [
            translate_particle_name_gate_to_geant4(k) for k in particle_keys_to_consider
        ]

        for particle in g4.G4ParticleTable.GetParticleTable().GetParticleList():
            add_step_limiter = False
            add_user_special_cuts = False
            p_name = str(particle.GetParticleName())

            if p_name in particles_to_consider:
                add_step_limiter = True
                add_user_special_cuts = True

            # this reproduces the logic of the Geant4's G4StepLimiterPhysics class
            if (
                ui.user_limits_particles["all_charged"] is True
                and particle.GetPDGCharge() != 0
            ):
                add_step_limiter = True

            if add_step_limiter is True or add_user_special_cuts is True:
                pm = particle.GetProcessManager()
                if add_step_limiter is True:
                    # G4StepLimiter for the max_step_size cut
                    g4_step_limiter = g4.G4StepLimiter("StepLimiter")
                    pm.AddDiscreteProcess(g4_step_limiter, 1)
                    # store limiter and cuts in lists to
                    # to avoid garbage collection after exiting the methods
                    self.g4_step_limiter_storage[p_name] = g4_step_limiter
                if add_user_special_cuts is True:
                    # G4UserSpecialCuts for the other cuts
                    g4_user_special_cuts = g4.G4UserSpecialCuts("UserSpecialCut")
                    pm.AddDiscreteProcess(g4_user_special_cuts, 1)
                    self.g4_special_user_cuts_storage[p_name] = g4_user_special_cuts


class Region(GateObject):
    """FIXME: Documentation of the Region class."""

    user_info_defaults = {}
    user_info_defaults["user_limits"] = (
        Box(
            {
                "max_step_size": None,
                "max_track_length": None,
                "min_ekine": None,
                "max_time": None,
                "min_range": None,
            }
        ),
        {
            "doc": "\tUser limits to be applied during tracking. \n"
            + "\tFIXME: Will be applied to all particles specified in the \n"
            + "\tlist under the `particles` keyword, if eligible.\n"
            + "\tUse `all` to apply tracking limits to all eligible particles.\n"
            + "\tThe following limits can be set:\n"
            + "\t* max_step_size\n"
            + "\t* max_track_length\n"
            + "\t* min_ekine\n"
            + "\t* max_time\n"
            + "\t* min_range\n",
            # expose_items=True means that the user_limits are also accessible directly
            # via Region.max_step_size, not only via Region.user_limits.max_step_size
            # that's more convenient for the user
            "expose_items": True,
        },
    )
    user_info_defaults["production_cuts"] = (
        Box(dict([(p, None) for p in cut_particle_names.keys()])),
        {
            "doc": "\tProduction cut per particle to be applied in volumes associated with this region.\n"
            + "\tShould be provided as key:value pair as: `particle_name` (string) : `cut_value` (numerical)\n"
            + "\tThe following particle names are allowed:\n"
            + "".join([f"\t* {p}\n" for p in cut_particle_names])
        },
    )
    user_info_defaults["em_switches"] = (
        Box([("deex", None), ("auger", None), ("pixe", None)]),
        {
            "doc": "Switch on/off EM parameters in this region. "
            "If None, the corresponding value from the world region is used.",
            "expose_items": True,
        },
    )

    def __init__(self, *args, **kwargs) -> None:
        # references to upper hierarchy level
        super().__init__(*args, **kwargs)

        self.physics_engine = None

        # dictionaries to hold volumes to which this region is associated
        # self.volumes = {}
        self.root_logical_volumes = {}

        # g4_objects; will be created by resp. initialize_XXX() methods
        self.g4_region = None
        self.g4_user_limits = None
        self.g4_production_cuts = None

        # flags for private use
        self._g4_region_initialized = False
        self._g4_user_limits_initialized = False
        self._g4_production_cuts_initialized = False

    @property
    def physics_manager(self):
        return self.simulation.physics_manager

    def reset(self):
        super().__init__(name=self.name, simulation=self.simulation)
        self.root_logical_volumes = {}

    # this version will work when Volume inherits from GateObject
    # def associate_volume(self, volume):
    #     volume_name = volume.name
    #     if volume_name not in self.root_logical_volumes.keys():
    #         self.root_logical_volumes[volume_name] = volume
    #     else:
    #         fatal(f'This volume {volume_name} is already associated with this region.')

    def close(self):
        self.release_g4_references()
        self.physics_engine = None

    def release_g4_references(self):
        self.g4_region = None
        self.g4_user_limits = None
        self.g4_production_cuts = None

    def to_dictionary(self):
        d = super().to_dictionary()
        d["root_logical_volumes_names"] = list(self.root_logical_volumes.keys())
        return d

    def from_dictionary(self, d):
        self.reset()
        super().from_dictionary(d)
        for vname in d["root_logical_volumes_names"]:
            self.associate_volume(vname)

    def need_step_limiter(self):
        if self.user_info["user_limits"]["max_step_size"] is not None:
            return True
        else:
            return False

    def need_user_special_cut(self):
        if (
            self.user_info["user_limits"]["max_track_length"] is not None
            or self.user_info["user_limits"]["min_ekine"] is not None
            or self.user_info["user_limits"]["max_time"] is not None
            or self.user_info["user_limits"]["min_range"] is not None
        ):
            return True
        else:
            return False

    @requires_fatal("physics_manager")
    def associate_volume(self, volume):
        # Allow volume object to be passed and retrieve its name in that case
        try:
            volume_name = volume.name
        except AttributeError:
            volume_name = volume

        if volume_name in self.root_logical_volumes:
            fatal(f"This volume {volume_name} is already associated with this region.")
        self.root_logical_volumes[volume_name] = None
        self.physics_manager.volumes_regions_lut[volume_name] = self

    def dump_production_cuts(self):
        s = ""
        for pname, cut in self.production_cuts.items():
            if cut is not None:
                s += f"{pname}: {cut}\n"
        return s

    @requires_fatal("physics_engine")
    def initialize(self):
        """
        This method wraps around all initialization methods of this class.

        It should be called from the physics_engine,
        after setting the self.physics_engine attribute.
        """
        self.initialize_volume_dictionaries()
        self.initialize_g4_production_cuts()
        self.initialize_g4_user_limits()
        self.initialize_g4_region()

    # This method is currently necessary because the actual volume objects
    # are only created at some point during initialization
    @requires_fatal("physics_engine")
    def initialize_volume_dictionaries(self):
        if self.physics_engine is None:
            fatal("No physics_engine defined.")
        for vname in self.root_logical_volumes.keys():
            self.root_logical_volumes[vname] = (
                self.physics_engine.simulation_engine.volume_engine.get_volume(vname)
            )

    def initialize_g4_region(self):
        if self._g4_region_initialized is True:
            fatal("g4_region already initialized.")

        rs = g4.G4RegionStore.GetInstance()
        self.g4_region = rs.FindOrCreateRegion(self.user_info.name)

        if self.g4_user_limits is not None:
            self.g4_region.SetUserLimits(self.g4_user_limits)

        # if self.g4_production_cuts is not None:
        self.g4_region.SetProductionCuts(self.g4_production_cuts)

        for vol in self.root_logical_volumes.values():
            self.g4_region.AddRootLogicalVolume(vol.g4_logical_volume, True)
            vol.g4_logical_volume.SetRegion(self.g4_region)

        self._g4_region_initialized = True

    def initialize_g4_production_cuts(self):
        self.user_info = Box(self.user_info)

        if self._g4_production_cuts_initialized is True:
            fatal("g4_production_cuts already initialized.")
        if self.g4_production_cuts is None:
            self.g4_production_cuts = g4.G4ProductionCuts()

        # 'all' overrides individual cuts per particle
        try:
            cut_for_all = self.user_info["production_cuts"]["all"]
        except KeyError:
            cut_for_all = None
        if cut_for_all is not None:
            for pname in self.user_info["production_cuts"].keys():
                if pname == "all":
                    continue
                g4_pname = translate_particle_name_gate_to_geant4(pname)
                self.g4_production_cuts.SetProductionCut(cut_for_all, g4_pname)
        else:
            for pname, cut in self.user_info["production_cuts"].items():
                if pname == "all":
                    continue
                # translate to G4 names, e.g. electron -> e+
                g4_pname = translate_particle_name_gate_to_geant4(pname)
                if cut is not None:
                    self.g4_production_cuts.SetProductionCut(cut, g4_pname)
                # If no cut is specified by user for this particle,
                # set it to the value specified for the world region
                else:
                    global_cut = self.physics_engine.g4_physics_list.GetCutValue(
                        g4_pname
                    )
                    self.g4_production_cuts.SetProductionCut(global_cut, g4_pname)

        self._g4_production_cuts_initialized = True

    def initialize_g4_user_limits(self):
        if self._g4_user_limits_initialized is True:
            fatal("g4_user_limits already initialized.")

        # check if any user limits have been set
        # if not, it is not necessary to create g4 objects
        if all([(ul is None) for ul in self.user_info["user_limits"].values()]) is True:
            self._g4_user_limits_initialized = True
            return

        self.g4_user_limits = g4.G4UserLimits()

        if self.user_info["user_limits"]["max_step_size"] is None:
            self.g4_user_limits.SetMaxAllowedStep(FLOAT_MAX)
        else:
            self.g4_user_limits.SetMaxAllowedStep(
                self.user_info["user_limits"]["max_step_size"]
            )

        if self.user_info["user_limits"]["max_track_length"] is None:
            self.g4_user_limits.SetUserMaxTrackLength(FLOAT_MAX)
        else:
            self.g4_user_limits.SetUserMaxTrackLength(
                self.user_info["user_limits"]["max_track_length"]
            )

        if self.user_info["user_limits"]["max_time"] is None:
            self.g4_user_limits.SetUserMaxTime(FLOAT_MAX)
        else:
            self.g4_user_limits.SetUserMaxTime(
                self.user_info["user_limits"]["max_time"]
            )

        if self.user_info["user_limits"]["min_ekine"] is None:
            self.g4_user_limits.SetUserMinEkine(0.0)
        else:
            self.g4_user_limits.SetUserMinEkine(
                self.user_info["user_limits"]["min_ekine"]
            )

        if self.user_info["user_limits"]["min_range"] is None:
            self.g4_user_limits.SetUserMinRange(0.0)
        else:
            self.g4_user_limits.SetUserMinRange(
                self.user_info["user_limits"]["min_range"]
            )

        self._g4_user_limits_initialized = True

    def initialize_em_switches(self):
        # if all switches are None, nothing is to be set
        if any([v is not None for v in self.em_switches.values()]):
            values_to_set = {}
            for k, v in self.em_switches.items():
                if v is None:  # try to recover switch from world
                    values_to_set[k] = self.physics_manager.em_switches_world[k]
                    if values_to_set[k] is None:
                        fatal(
                            f"No value (True/False) provided for em_switch {k} in region {self.name} and no corresponding value set for the world either."
                        )
                else:
                    values_to_set[k] = v
            self.physics_engine.g4_em_parameters.SetDeexActiveRegion(
                self.name,
                values_to_set["deex"],
                values_to_set["auger"],
                values_to_set["pixe"],
            )


def get_enum_values(enum_class):
    # Filter out special Python attributes, methods, and pybind11 specific attributes
    return list(enum_class.__members__.keys())
    # return [
    #     attr
    #     for attr in dir(enum_class)
    #     if not attr.startswith("__")
    #        and not callable(getattr(enum_class, attr))
    #        and attr not in ["name", "value"]
    # ]


def load_optical_surface_properties_from_xml(surface_properties_file, surface_name):
    """
    This function extracts the information related to multiple surfaces
    from SurfaceProperties.xml
    """

    try:
        xml_tree = ET.parse(surface_properties_file)
    except FileNotFoundError:
        fatal(
            f"Could not find the surface_optical_properties_file {surface_properties_file}."
        )
    xml_root = xml_tree.getroot()

    found_surface_names = set()
    surface_properties = None
    for m in xml_root.findall("surface"):
        if m.get("name") == surface_name:
            surface_properties = {
                "base_properties": {
                    "surface_model": m.get("model"),
                    "surface_name": surface_name,
                    "surface_type": m.get("type"),
                    "surface_finish": m.get("finish"),
                    "surface_sigma_alpha": m.get("sigmaalpha"),
                },
                "constant_properties": {},
                "vector_properties": {},
            }

            # Handle propertyvector elements for UNIFIED Model
            for ptable in m.findall("propertiestable"):
                for prop_vector in ptable.findall("propertyvector"):
                    prop_vector_name = prop_vector.get("name")
                    prop_vector_value_unit = prop_vector.get("unit")
                    prop_vector_energy_unit = prop_vector.get("energyunit")

                    if prop_vector_value_unit is not None:
                        value_unit = g4_units[prop_vector_value_unit]
                    else:
                        value_unit = 1.0

                    if prop_vector_energy_unit is not None:
                        energy_unit = g4_units[prop_vector_energy_unit]
                    else:
                        energy_unit = 1.0

                    # Handle ve elements inside propertyvector
                    ve_energy_list = []
                    ve_value_list = []

                    for ve in prop_vector.findall("ve"):
                        ve_energy_list.append(float(ve.get("energy")) * energy_unit)
                        ve_value_list.append(float(ve.get("value")) * value_unit)

                    surface_properties["vector_properties"][prop_vector_name] = {
                        "prop_vector_value_unit": prop_vector_value_unit,
                        "prop_vector_energy_unit": prop_vector_energy_unit,
                        "ve_energy_list": ve_energy_list,
                        "ve_value_list": ve_value_list,
                    }

    if surface_properties is not None:
        return surface_properties
    else:
        fatal(
            f"No surface named {surface_name} not found in the XML file {surface_properties_file}"
        )


def load_optical_properties_from_xml(optical_properties_file, material_name):
    """This function parses an xml file containing optical material properties.
    Fetches property elements and property vector elements.

    Returns a dictionary with the properties or None if the material is not found in the file.
    """
    try:
        xml_tree = ET.parse(optical_properties_file)
    except FileNotFoundError:
        fatal(f"Could not find the optical_properties_file {optical_properties_file}.")
    xml_root = xml_tree.getroot()

    xml_entry_material = None
    for m in xml_root.findall("material"):
        # FIXME: some names might follow different conventions, e.g. 'Water' vs. 'G4_WATER'
        # using variants of the name is a possible solution, but this should be reviewed
        if str(m.get("name")) in get_material_name_variants(material_name):
            xml_entry_material = m
            break
    if xml_entry_material is None:
        warning(
            f"Could not find any optical material properties for material {material_name} "
            f"in file {optical_properties_file}."
        )
        return

    material_properties = {"constant_properties": {}, "vector_properties": {}}

    for ptable in xml_entry_material.findall("propertiestable"):
        # Handle property elements in XML document
        for prop in ptable.findall("property"):
            property_name = prop.get("name")
            property_value = float(prop.get("value"))
            property_unit = prop.get("unit")

            # apply unit if applicable
            if property_unit is not None:
                if len(property_unit.split("/")) == 2:
                    unit = property_unit.split("/")[1]
                else:
                    unit = property_unit
                property_value *= g4_units[unit]

            material_properties["constant_properties"][property_name] = {
                "property_value": property_value,
                "property_unit": property_unit,
            }

        # Handle propertyvector elements
        for prop_vector in ptable.findall("propertyvector"):
            prop_vector_name = prop_vector.get("name")
            prop_vector_value_unit = prop_vector.get("unit")
            prop_vector_energy_unit = prop_vector.get("energyunit")

            if prop_vector_value_unit is not None:
                value_unit = g4_units[prop_vector_value_unit]
            else:
                value_unit = 1.0

            if prop_vector_energy_unit is not None:
                energy_unit = g4_units[prop_vector.get("energyunit")]
            else:
                energy_unit = 1.0

            # Handle ve elements inside propertyvector
            ve_energy_list = []
            ve_value_list = []
            for ve in prop_vector.findall("ve"):
                ve_energy_list.append(float(ve.get("energy")) * energy_unit)
                ve_value_list.append(float(ve.get("value")) * value_unit)

            material_properties["vector_properties"][prop_vector_name] = {
                "prop_vector_value_unit": prop_vector_value_unit,
                "prop_vector_energy_unit": prop_vector_energy_unit,
                "ve_energy_list": ve_energy_list,
                "ve_value_list": ve_value_list,
            }

    return material_properties


def create_g4_optical_properties_table(material_properties_dictionary):
    """Creates and fills a G4MaterialPropertiesTable with values from a dictionary created by a parsing function,
    e.g. from an xml file.
    Returns G4MaterialPropertiesTable.
    """

    g4_material_table = g4.G4MaterialPropertiesTable()

    for property_name, data in material_properties_dictionary[
        "constant_properties"
    ].items():
        # check whether the property is already present
        create_new_key = (
            property_name not in g4_material_table.GetMaterialConstPropertyNames()
        )
        if create_new_key is True:
            warning(
                f"Found property {property_name} in optical properties file which is not known to Geant4. "
                f"I will create the property for you, but you should verify whether physics are correctly modeled."
            )
        g4_material_table.AddConstProperty(
            g4.G4String(property_name), data["property_value"], create_new_key
        )

    for property_name, data in material_properties_dictionary[
        "vector_properties"
    ].items():
        # check whether the property is already present
        create_new_key = (
            property_name not in g4_material_table.GetMaterialPropertyNames()
        )
        if create_new_key is True:
            warning(
                f"Found property {property_name} in optical properties file which is not known to Geant4. "
                f"I will create the property for you, but you should verify whether physics are correctly modeled."
            )
        g4_material_table.AddProperty(
            g4.G4String(property_name),
            data["ve_energy_list"],
            data["ve_value_list"],
            create_new_key,
            False,
        )

    return g4_material_table


class OpticalSurface(GateObject):
    """
    Class used to create an Optical Surface between two volumes

    G4OpticalSurface is used to create an optical surface

    G4LogicalBorderSurface is used to assign the optical surface
    between two volumes.
    """

    user_info_defaults = {
        "volume_from": (
            None,
            {
                "doc": "The volume from which photons propagate through the optical surface. "
            },
        ),
        "volume_to": (
            None,
            {
                "doc": "The volume into which the photons propagate coming from the surface. "
            },
        ),
        "g4_surface_name": (
            None,
            {
                "doc": "Name of the Geant4 surface to be created between volume_from and volume_to"
            },
        ),
    }

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.physics_engine = None

        # dictionary holding optical surface properties
        # populate from information stored in an external file
        # whose location is specified via physics_manager.surface_properties_file
        self.optical_surface_properties_dict = None

        # Store Geant4 Optical Surface object
        self.g4_optical_surface = None
        # Store Geant4 Logical Border Surface object
        self.g4_logical_border_surface = None
        # Store Geant4 object for material properties table
        self.g4_optical_surface_table = None

    # shortcut for convenience
    @property
    def physics_manager(self):
        if self.simulation is not None:
            return self.simulation.physics_manager
        else:
            return None

    def release_g4_references(self):
        self.g4_optical_surface = None
        self.g4_logical_border_surface = None
        self.g4_optical_surface_table = None

    def close(self):
        self.release_g4_references()
        self.physics_engine = None
        super().close()

    def reset(self):
        self.__init__(name=self.name, physics_manager=self.physics_manager)

    def __getstate__(self):
        return_dict = super().__getstate__()
        return_dict["g4_optical_surface"] = None
        return_dict["g4_logical_border_surface"] = None
        return_dict["g4_optical_surface_table"] = None
        return return_dict

    @requires_fatal("physics_engine")
    def initialize(self):
        # Create object of Geant4 Optical Surface
        self.g4_optical_surface = g4.G4OpticalSurface(g4.G4String(self.g4_surface_name))

        self.optical_surface_properties_dict = load_optical_surface_properties_from_xml(
            self.physics_manager.surface_properties_file,
            self.g4_surface_name,
        )

        # Set properties to create G4 Optical Surface object
        surface_base_properties = self.optical_surface_properties_dict[
            "base_properties"
        ]

        # Set model (eg. Unified, LUT_Davis)
        model_name = surface_base_properties["surface_model"]
        try:
            model = getattr(g4.G4OpticalSurfaceModel, model_name)
            self.g4_optical_surface.SetModel(model)
        except AttributeError:
            fatal(
                f"Unknown Model - {model_name} \n"
                f"Available models are {get_enum_values(g4.G4OpticalSurfaceModel)}"
            )

        # Set surface type
        surface_type_name = surface_base_properties["surface_type"]
        try:
            surface_type = getattr(g4.G4SurfaceType, surface_type_name)
            self.g4_optical_surface.SetType(surface_type)
        except AttributeError:
            fatal(
                f"Unknown Surface Type - {surface_type_name} \n"
                f"Available Surface Types are {get_enum_values(g4.G4SurfaceType)}"
            )

        # Set finish
        surface_finish_name = surface_base_properties["surface_finish"]
        try:
            surface_finish = getattr(
                g4.G4OpticalSurfaceFinish, surface_finish_name, None
            )
            self.g4_optical_surface.SetFinish(surface_finish)
        except AttributeError:
            fatal(
                f"Unknown Surface Finish - {surface_finish_name} \n"
                f"Available Surface Finishes are {get_enum_values(g4.G4OpticalSurfaceFinish)}"
            )

        # Set sigma alpha
        surface_sigma_alpha = surface_base_properties["surface_sigma_alpha"]

        if surface_sigma_alpha is not None:
            self.g4_optical_surface.SetSigmaAlpha(
                float(surface_sigma_alpha) * g4_units.deg
            )

        # Set surface properties table
        self.g4_optical_surface_table = create_g4_optical_properties_table(
            self.optical_surface_properties_dict
        )

        self.g4_optical_surface.SetMaterialPropertiesTable(
            self.g4_optical_surface_table
        )

        # Set the Optical Surface between two volumes
        # g4_physical_volumes (local variables are OK because
        # permanent references are stored inside the respective python Volume instances)
        g4_physical_volume_from = (
            self.physics_engine.simulation_engine.volume_engine.get_volume(
                self.volume_from
            ).get_g4_physical_volume(0)
        )

        g4_physical_volume_to = (
            self.physics_engine.simulation_engine.volume_engine.get_volume(
                self.volume_to
            ).get_g4_physical_volume(0)
        )

        self.g4_logical_border_surface = g4.G4LogicalBorderSurface(
            g4.G4String(self.g4_surface_name),
            g4_physical_volume_from,
            g4_physical_volume_to,
            self.g4_optical_surface,
        )


process_cls(Region)
process_cls(OpticalSurface)



from .utility import create_hard_links_to_nodes_in_group
from ..exception import fatal
from .base import ProcessingGroupBase
from .unitbase import ProcessingUnitBase


class ProcessingSequence(ProcessingGroupBase, ProcessingUnitBase):
    """A special kind of processing group that has only one branch,
    i.e. a single root and each node has a single child only.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def update_processing_tree(self):
        super().update_processing_tree()
        tree_roots = self.tree_roots
        if len(tree_roots) > 1:
            fatal(f"Processing sequence {self.name} has more than one tree root. ")
        if len(list(tree_roots)[0].leaves) > 1:
            fatal(f"Processing sequence {self.name} has more than one leaf. ")

    def get_unit_output_group(self):
        return self.get_output_groups()[0]

    def do_your_job(self):
        self.tree_roots[0].update_output()
        for k, v in self.tree_roots[0].output_data_handles.items():
            self.output_data_handles[k] = v
        create_hard_links_to_nodes_in_group(
            self.hdf5_group, self.get_unit_output_group()
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from opengate.exception import GateDeprecationError
from opengate.tests.utility import (
    get_default_test_paths,
    test_ok,
    print_test,
)
from opengate.utility import g4_units
import opengate as gate


def set_wrong_attribute(obj, attr):
    # set a WRONG attribute
    number_of_warnings_before = getattr(obj, "number_of_warnings")
    setattr(obj, attr, "nothing")
    number_of_warnings_after = getattr(obj, "number_of_warnings")

    # check the number of warnings before and after
    print(
        f"Number of warnings for {obj.type_name} object {obj.name}: {number_of_warnings_before}"
    )
    print(
        f"Number of warnings for {obj.type_name} object {obj.name}: {number_of_warnings_after}"
    )
    b = number_of_warnings_after - number_of_warnings_before == 1
    print_test(
        b, f"Tried to set a wrong attribute '{attr}'. It should print a single warning"
    )
    return b


if __name__ == "__main__":
    paths = get_default_test_paths(
        __file__,
    )

    sim = gate.Simulation()

    m = g4_units.m
    cm = g4_units.cm
    keV = g4_units.keV
    mm = g4_units.mm
    um = g4_units.um
    Bq = g4_units.Bq

    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    is_ok = True

    # create the volume without adding it to the simulation
    waterbox = gate.geometry.volumes.BoxVolume(name="Waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"
    # provoke a warning
    is_ok = is_ok and set_wrong_attribute(waterbox, "mohter")
    # now add the volume to the simulation
    sim.add_volume(waterbox)
    # a warning about the wrong attribute "mohter" should appear
    # in the list of warnings at the end of the simulation
    # because it should be transferred from the object's warning cache
    # to the simulation's warning cache once the object is added to the simulation

    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 20

    # test wrong attributes
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True
    try:
        stats.mother = "nothing"
        is_ok = False
    except GateDeprecationError as e:
        print("Exception caught: ", e)
        pass
    print_test(
        is_ok, f"Try to set deprecated attribute 'mother', it should raise an exception"
    )
    print()

    is_ok = is_ok and set_wrong_attribute(stats, "TOTO")
    is_ok = is_ok and set_wrong_attribute(sim, "nthreads")

    sim.run(start_new_process=True)

    found_warning_about_waterbox_mohter = False
    print(sim.warnings)
    for w in sim.warnings:
        if "mohter" in w:
            found_warning_about_waterbox_mohter = True
            break
    is_ok = is_ok and found_warning_about_waterbox_mohter
    print(
        f"(after run) Number of warnings for stats object: {stats.number_of_warnings}"
    )

    b = stats.number_of_warnings == 1
    print_test(b, f"No additional warning should be raised")
    is_ok = is_ok and b

    print(is_ok)
    test_ok(is_ok)
